from typing import List, Optional

from pygls.workspace import Document, Workspace
from galaxyls.services.tools.common import TestsDiscoveryService
from galaxyls.services.tools.document import GalaxyToolXmlDocument
from galaxyls.services.xml.parser import XmlDocumentParser
from galaxyls.services.xml.utils import convert_document_offset_to_line
from galaxyls.types import TestInfoResult, TestSuiteInfoResult
from galaxyls.services.validation import DocumentValidator


class ToolTestsDiscoveryService(TestsDiscoveryService):
    document_validator = DocumentValidator()

    def discover_tests_in_workspace(self, workspace: Workspace) -> List[TestSuiteInfoResult]:
        rval: List[TestSuiteInfoResult] = []
        for doc_uri in workspace.documents:
            document = workspace.get_document(doc_uri)
            if self.document_validator.is_tool_document(document):
                test_suite = self._get_test_suite_from_document(document)
                if test_suite:
                    rval.append(test_suite)
        return rval

    def _get_test_suite_from_document(self, document: Document) -> Optional[TestSuiteInfoResult]:
        xml_document = XmlDocumentParser().parse(document)
        tool = GalaxyToolXmlDocument.from_xml_document(xml_document)
        tool_id = tool.get_tool_id()
        if tool_id:
            tests = tool.get_tests()
            suite_tests: List[TestInfoResult] = []
            id = 1
            for test in tests:
                line = convert_document_offset_to_line(xml_document.document, test.start_tag_open_offset)
                suite_tests.append(
                    TestInfoResult(
                        tool_id=tool_id,
                        test_id=str(id),
                        file=xml_document.document.path,
                        line=line,
                    ),
                )
                id += 1
            return TestSuiteInfoResult(
                tool_id=tool_id,
                file=xml_document.document.path,
                children=suite_tests,
            )



from typing import List, Optional, cast

from anytree import NodeMixin, find
from galaxyls.services.tools.constants import (
    BOOLEAN,
    CONDITIONAL,
    FALSEVALUE,
    INPUTS,
    MIN,
    NAME,
    OPTION,
    PARAM,
    REPEAT,
    SECTION,
    SELECT,
    TRUEVALUE,
    TYPE,
    VALUE,
    WHEN,
)
from galaxyls.services.xml.nodes import XmlElement


class InputNode(NodeMixin):
    """Represents a logical input node in the tool wrapper file.

    Each node contains references to direct children classified by their type."""

    def __init__(self, name: str, element: Optional[XmlElement] = None, parent: NodeMixin = None):
        self.name: str = name
        self.element: Optional[XmlElement] = element
        self.parent = parent
        self.params: List[XmlElement] = []
        self.repeats: List[RepeatInputNode] = []
        self.sections: List[SectionInputNode] = []

    def __repr__(self) -> str:
        return self.name


class ConditionalInputNode(InputNode):
    """Represents a conditional input branch node in the tool wrapper file.

    The 'option_param' field contains the first select or boolean param and it's
    value is set to the 'option' field of one of the possible 'when' definitions.
    """

    def __init__(self, name: str, option: str, element: Optional[XmlElement] = None, parent: InputNode = None):
        super().__init__(name, element, parent)
        self.option_param: XmlElement = element.elements[0]
        self.option: str = option
        self.options: List[str] = self._get_options()

    def __repr__(self) -> str:
        return f"{self.name} - {self.option}"

    @property
    def is_first_option(self) -> bool:
        if self.options:
            return self.option == self.options[0]
        return True

    @property
    def is_last_option(self) -> bool:
        if self.options:
            return self.option == self.options[-1]
        return True

    def _get_options(self) -> List[str]:
        try:
            type_attr = self.option_param.get_attribute(TYPE)
            if type_attr == SELECT:
                options = self.option_param.get_children_with_name(OPTION)
                return list(filter(None, [option.get_attribute(VALUE) for option in options]))
            elif type_attr == BOOLEAN:
                options = []
                true_value = self.option_param.get_attribute(TRUEVALUE)
                if true_value:
                    options.append(true_value)
                false_value = self.option_param.get_attribute(FALSEVALUE)
                if false_value:
                    options.append(false_value)
                return options
            return []
        except BaseException:
            return []


class RepeatInputNode(InputNode):
    """Represents an input node that will be repeated 'min' times."""

    def __init__(self, name: str, min: int, element: Optional[XmlElement] = None):
        super().__init__(name, element, parent=None)
        self.min: int = min


class SectionInputNode(InputNode):
    """Represents a section input node which is used to group other nodes."""

    def __init__(self, name: str, element: Optional[XmlElement] = None):
        super().__init__(name, element, parent=None)


class GalaxyToolInputTree:
    """The branches of this tree contains all the inputs within a conditional path for a specific option."""

    def __init__(self, inputs: Optional[XmlElement] = None) -> None:
        self._root: InputNode = InputNode(INPUTS, inputs)
        if inputs:
            self._build_input_tree(inputs, self._root)

    @property
    def leaves(self) -> List[InputNode]:
        """The leaves of the tree contain all the final elements of the conditional branches."""
        return list(self._root.leaves)

    def _build_input_tree(self, inputs: XmlElement, parent: InputNode) -> None:
        """Given the XML element containing the inputs of the tool, this method recursively builds
        an expanded input tree.

        Args:
            inputs (XmlElement): The XML element corresponding to the inputs.
            parent (InputNode): The node that will hold all the inputs.
        """
        parent.params = inputs.get_children_with_name(PARAM)

        conditionals = inputs.get_children_with_name(CONDITIONAL)
        for conditional in conditionals:
            self._build_conditional_input_tree(conditional, parent)

        repeats = inputs.get_children_with_name(REPEAT)
        for repeat in repeats:
            repeat_node = self._build_repeat_input_tree(repeat)
            if repeat_node:
                parent.repeats.append(repeat_node)

        sections = inputs.get_children_with_name(SECTION)
        for section in sections:
            section_node = self._build_section_input_tree(section)
            if section_node:
                parent.sections.append(section_node)

    def _build_conditional_input_tree(self, conditional: XmlElement, parent: InputNode) -> None:
        """Populates the 'parent' node with a ConditionalInputNode for each of the declared options
        in the given 'conditional' XmlElement.

        Args:
            conditional (XmlElement): The XML element (conditional tag).
            parent (InputNode): The InputNode that will hold the conditional branches and it's elements.
        """
        param = conditional.elements[0]  # first child must be select or boolean
        if param.get_attribute(TYPE) == SELECT:
            options = param.get_children_with_name(OPTION)
            for option in options:
                option_value = option.get_attribute(VALUE)
                self._build_conditional_option_branch(conditional, parent, option_value)
        elif param.get_attribute(TYPE) == BOOLEAN:
            true_value = param.get_attribute(TRUEVALUE)
            if true_value:
                self._build_conditional_option_branch(conditional, parent, true_value)
            false_value = param.get_attribute(FALSEVALUE)
            if false_value:
                self._build_conditional_option_branch(conditional, parent, false_value)

    def _build_conditional_option_branch(
        self, conditional: XmlElement, parent: InputNode, option_value: Optional[str] = None
    ) -> None:
        """Builds a conditional branch in the input tree with the given 'option_value'.

        Args:
            conditional (XmlElement): The <conditional> XML element.
            parent (InputNode): The input node that will contain this branch.
            option_value (str): The value of the option selected in this conditional branch.
        """
        name = conditional.get_attribute(NAME)
        if name and option_value:
            conditional_node = ConditionalInputNode(name, option_value, element=conditional, parent=parent)
            when = find(
                conditional, filter_=lambda el: el.name == WHEN and el.get_attribute(VALUE) == option_value, maxlevel=2
            )
            when = cast(XmlElement, when)
            if when:
                self._build_input_tree(when, conditional_node)

    def _build_repeat_input_tree(self, repeat: XmlElement) -> Optional[RepeatInputNode]:
        """Builds and returns a RepeatInputNode from a 'repeat' XML tag with the minimum number
        of repetitions defined.

        Args:
            repeat (XmlElement): The XML repeat tag.

        Returns:
            Optional[RepeatInputNode]: The resulting RepeatInputNode with it's own input children.
        """
        name = repeat.get_attribute(NAME)
        if name:
            min = 1
            min_attr = repeat.get_attribute(MIN)
            if min_attr and min_attr.isdigit:
                min = int(min_attr)
            repeat_node = RepeatInputNode(name, min, repeat)
            self._build_input_tree(repeat, repeat_node)
            return repeat_node
        return None

    def _build_section_input_tree(self, section: XmlElement) -> Optional[SectionInputNode]:
        """Builds and returns a SectionInputNode from a 'section' XML tag.

        Args:
            section (XmlElement): The XML section tag.

        Returns:
            Optional[SectionInputNode]: The resulting SectionInputNode with it's own input children.
        """
        name = section.get_attribute(NAME)
        if name:
            section_node = SectionInputNode(name, section)
            self._build_input_tree(section, section_node)
            return section_node
        return None



"""Utilities to validate Galaxy xml tool wrappers.
"""

from pathlib import Path
from typing import List, Optional

from lxml import etree
from pygls.lsp.types import Diagnostic, DiagnosticRelatedInformation, Location, Position, Range
from pygls.workspace import Document

from galaxy.util import xml_macros
from galaxyls.constants import DiagnosticCodes
from galaxyls.services.macros import remove_macros
from galaxyls.services.tools.document import GalaxyToolXmlDocument
from galaxyls.services.xml.document import XmlDocument

EXPAND_DOCUMENT_URI_SUFFIX = "%20%28Expanded%29"


class GalaxyToolValidationService:
    """Service providing diagnostics for errors in the XML validation."""

    def __init__(self, server_name: str, xsd_schema: etree.XMLSchema):
        """Initializes the validator"""
        self.server_name = server_name
        self.xsd_schema = xsd_schema

    def validate_document(self, xml_document: XmlDocument) -> List[Diagnostic]:
        """Validates the XML document and returns a list of diagnostics
        if there are any problems.

        Args:
            xml_document (XmlDocument): The XML document. Can be a tool wrapper or macro
            definition file.

        Returns:
            List[Diagnostic]: The list of issues found in the document.
        """

        syntax_errors = self._check_syntax(xml_document.document)
        if syntax_errors:
            return syntax_errors

        if not xml_document.is_tool_file:
            return []
        tool = GalaxyToolXmlDocument.from_xml_document(xml_document)
        try:
            xml_tree = etree.fromstring(tool.source)
            return self._validate_tree(xml_tree)
        except ExpandMacrosFoundException:
            result = self._validate_expanded(tool)
            return result
        except etree.XMLSyntaxError as e:
            return self._build_diagnostics_from_syntax_error(e)

    def _check_syntax(self, document: Document) -> List[Diagnostic]:
        """Check if the XML document contains any syntax error and returns it in a list.

        Args:
            document (Document): The XML document.

        Returns:
            List[Diagnostic]: The list containing the syntax error found or an empty list.
        """
        try:
            etree.fromstring(document.source)
            return []
        except etree.XMLSyntaxError as e:
            return self._build_diagnostics_from_syntax_error(e)

    def _validate_expanded(self, tool: GalaxyToolXmlDocument) -> List[Diagnostic]:
        """Validates the document after loading all the macros referenced and expands them.

        Args:
            document (Document): [description]
            xml_tree (etree.ElementTree): [description]

        Returns:
            List[Diagnostic]: [description]
        """
        try:
            expanded_tool_tree, _ = xml_macros.load_with_references(tool.path)
            expanded_xml = remove_macros(expanded_tool_tree)
            root = expanded_xml.getroot()
            self.xsd_schema.assertValid(root)
            return []
        except etree.DocumentInvalid as e:
            diagnostics = self._build_diagnostics_for_expanded_macros(tool, e)
            return diagnostics

        except etree.XMLSyntaxError as e:
            return self._build_diagnostics_for_macros_file_syntax_error(tool, e)

    def _validate_tree(self, xml_tree: etree._ElementTree) -> List[Diagnostic]:
        """Validates an XML tree against the XSD schema.

        Args:
            xml_tree (etree.ElementTree): The root element

        Returns:
            List[Diagnostic]: The diagnosis results
        """
        try:
            self.xsd_schema.assertValid(xml_tree)
            return []
        except etree.DocumentInvalid as e:
            return self._build_diagnostics(e.error_log, xml_tree)

    def _build_diagnostics(
        self, error_log: etree._ListErrorLog, xml_tree: Optional[etree._ElementTree] = None
    ) -> List[Diagnostic]:
        """Gets a list of diagnostics from the XSD validation error log.

        Args:
            error_log (etree._ListErrorLog): The error log generated after XSD schema validation.
            xml_tree (Optional[etree.ElementTree], optional): The element tree associated with
            the error log. Defaults to None.

        Raises:
            ExpandMacrosFoundException: This is raised when a macro ``expand`` element is found in the error log.
            The ``expand`` element is not part of the XSD so, when this exception is raised, we know that
            the document must expand the macros before validation.

        Returns:
            List[Diagnostic]: The list of resulting diagnostic items found in the error log.
        """
        diagnostics = []
        for error in error_log.filter_from_errors():
            if "expand" in error.message:
                raise ExpandMacrosFoundException(xml_tree)

            result = Diagnostic(
                range=Range(
                    start=Position(line=error.line - 1, character=error.column),
                    end=Position(line=error.line - 1, character=error.column),
                ),
                message=error.message,
                source=self.server_name,
            )
            diagnostics.append(result)
        return diagnostics

    def _build_diagnostics_from_syntax_error(self, error: etree.XMLSyntaxError) -> List[Diagnostic]:
        """Builds a Diagnostic element from a XMLSyntaxError.

        Args:
            error (etree.XMLSyntaxError): The syntax error.

        Returns:
            Diagnostic: The converted Diagnostic item.
        """
        result = Diagnostic(
            range=Range(
                start=Position(line=error.lineno - 1, character=error.position[0] - 1),
                end=Position(line=error.lineno - 1, character=error.position[1] - 1),
            ),
            message=error.msg,
            source=self.server_name,
        )
        return [result]

    def _build_diagnostics_for_macros_file_syntax_error(
        self, tool: GalaxyToolXmlDocument, e: etree.XMLSyntaxError
    ) -> List[Diagnostic]:
        result = Diagnostic(
            range=tool.get_import_macro_file_range(e.filename),
            message=e.msg,
            source=self.server_name,
            related_information=[
                DiagnosticRelatedInformation(
                    message="Syntax error found on imported file.",
                    location=Location(
                        uri=Path(e.filename).as_uri(),
                        range=Range(
                            start=Position(line=e.lineno - 1, character=e.offset),
                            end=Position(line=e.lineno - 1, character=e.offset),
                        ),
                    ),
                )
            ],
        )
        return [result]

    def _build_diagnostics_for_expanded_macros(
        self, tool: GalaxyToolXmlDocument, e: etree.DocumentInvalid
    ) -> List[Diagnostic]:
        virtual_uri = tool.xml_document.document.uri.replace("file", "gls-expand")
        diagnostics = [
            Diagnostic(
                range=tool.get_macros_range(),
                message=error.message,
                source=self.server_name,
                code=DiagnosticCodes.INVALID_EXPANDED_TOOL,
                related_information=[
                    DiagnosticRelatedInformation(
                        message=(
                            "The validation error ocurred on the expanded version of "
                            "the document, i.e. after replacing macros. "
                            "Click here to preview the expanded document."
                        ),
                        location=Location(
                            uri=f"{virtual_uri}{EXPAND_DOCUMENT_URI_SUFFIX}",
                            range=Range(
                                start=Position(line=error.line, character=error.column),
                                end=Position(line=error.line, character=error.column),
                            ),
                        ),
                    )
                ],
            )
            for error in e.error_log.filter_from_errors()
        ]
        return diagnostics


class ExpandMacrosFoundException(Exception):
    """This exceptions indicates that the current tool contains
    macros and they should be expanded before validation.
    """



from typing import List, Optional

from galaxyls.services.tools.common import ToolParamAttributeSorter
from galaxyls.services.tools.constants import (
    ARGUMENT,
    CHECKED,
    FALSEVALUE,
    FORMAT,
    HELP,
    LABEL,
    MAX,
    MIN,
    NAME,
    OPTIONAL,
    PARAM,
    TRUEVALUE,
    TYPE,
    VALUE,
)
from galaxyls.services.xml.document import XmlDocument
from galaxyls.services.xml.nodes import XmlAttribute, XmlElement
from galaxyls.services.xml.utils import convert_document_offsets_to_range
from galaxyls.types import ReplaceTextRangeResult

ORDER_LAST = 100


class IUCToolParamAttributeSorter(ToolParamAttributeSorter):
    IUC_PARAM_ATTR_ORDER = {
        NAME: 0,
        ARGUMENT: 1,
        TYPE: 2,
        FORMAT: 3,
        MIN: 4,
        MAX: 5,
        VALUE: 6,
        TRUEVALUE: 4,
        FALSEVALUE: 5,
        CHECKED: 6,
        OPTIONAL: 7,
        LABEL: 8,
        HELP: 9,
    }

    def sort_param_attributes(self, param: XmlElement, xml_document: XmlDocument) -> Optional[ReplaceTextRangeResult]:
        if param and param.name == PARAM and param.has_attributes:
            attribute_names = param.get_attribute_names()
            sorted_attribute_names = self._sort_attribute_names(attribute_names)
            if attribute_names == sorted_attribute_names:
                return None
            sorted_attributes_text = self._get_param_attributes_as_text_sorted(param, sorted_attribute_names)
            start, end = param.get_attributes_offsets()
            return ReplaceTextRangeResult(
                replace_range=convert_document_offsets_to_range(xml_document.document, start, end),
                text=sorted_attributes_text,
            )
        return None

    def sort_document_param_attributes(self, xml_document: XmlDocument) -> List[ReplaceTextRangeResult]:
        params = xml_document.find_all_elements_with_name(PARAM)
        rval: List[ReplaceTextRangeResult] = []
        for param in params:
            result = self.sort_param_attributes(param, xml_document)
            if result:
                rval.append(result)
        return rval

    def _sort_attribute_names(self, attributes: List[str]) -> List[str]:
        return sorted(attributes, key=lambda attr: self._get_attr_order(attr))

    def _get_attr_order(self, attr_name: str) -> int:
        return self.IUC_PARAM_ATTR_ORDER.get(attr_name, ORDER_LAST)

    def _get_param_attributes_as_text_sorted(self, element: XmlElement, sorted_attr_names: List[str]) -> str:
        printed_attributes = [self._attr_to_text(element.attributes[attr_name]) for attr_name in sorted_attr_names]
        return " ".join(printed_attributes)

    def _attr_to_text(self, attr: XmlAttribute) -> str:
        if attr and attr.name and attr.value:
            return f'{attr.name}="{attr.get_value()}"'
        return ""



"""This module contains shared types across all modules.
"""

from typing import List, Optional
from pydantic.main import BaseModel
from pygls.lsp.types import Position, Range


class AutoCloseTagResult:
    """Contains the code snippet and the range in the text document that
    will be returned to the client when a tag auto-close is requested.
    """

    def __init__(self, snippet: str, replace_range: Optional[Range] = None):
        self.snippet = snippet
        self.range = replace_range


class GeneratedSnippetResult:
    """Contains the auto-generated code snippet and the position in the text document
    where it should be inserted.

    Optionally, a Range can be provided to insert the snippet replacing the
    text inside the given range.
    """

    def __init__(
        self,
        snippet: str,
        insert_position: Position,
        replace_range: Optional[Range] = None,
        error_message: Optional[str] = None,
    ):
        self.snippet = snippet
        self.position = insert_position
        self.replace_range = replace_range
        self.error_message = error_message

    @staticmethod
    def as_error(error_message: str) -> "GeneratedSnippetResult":
        """Returns a GeneratedSnippetResult with empty values and the given error message.

        The error message should be displayed or logged by the client."""
        return GeneratedSnippetResult("", Position(line=0, character=0), error_message=error_message)


class ReplaceTextRangeResult:
    """Represents a position range in the document that should be replaced with some text."""

    def __init__(self, replace_range: Range, text: str) -> None:
        self.replace_range = replace_range
        self.text = text


class TestInfoResult:
    """Contains information about a particular test case."""

    def __init__(self, tool_id: str, test_id: str, file: str, line: int, skipped: bool = False) -> None:
        self.id = f"{tool_id}:{test_id}"
        self.label = f"{tool_id}:Test #{test_id}"
        self.file = file
        self.line = line
        self.skipped = skipped
        self.description: Optional[str] = None
        self.tooltip: Optional[str] = None
        self.debuggable: bool = False
        self.errored: bool = False
        self.message: Optional[str] = None


class TestSuiteInfoResult:
    """Contains information about all the tests for a tool wrapper."""

    def __init__(self, tool_id: str, file: str, children: List[TestInfoResult] = None) -> None:
        self.id = tool_id
        self.label = f"{tool_id} tests"
        self.file = file
        self.children = children
        self.line: int = 0
        self.debuggable: bool = False
        self.description: Optional[str] = None
        self.tooltip: Optional[str] = None

        self.errored: bool = False
        self.message: Optional[str] = None


class GeneratedExpandedDocument(BaseModel):
    """Represents a tool document with all the macros expanded."""

    content: Optional[str]
    error_message: Optional[str]



""" This code is based on the Eclipse/Lemminx XML language server implementation:
https://github.com/eclipse/lemminx/tree/master/org.eclipse.lemminx/src/main/java/org/eclipse/lemminx/dom

Only the minimum subset of the XML dialect used by Galaxy tool wrappers is supported.
"""

from typing import Callable, List

from pygls.lsp.types import Position, Range
from pygls.workspace import Document

from .constants import NEW_LINE, _LAN, WHITESPACE_CHARS


def convert_document_offset_to_line(document: Document, offset: int) -> int:
    """Converts the given offset in the document to the corresponding line.

    Args:
        document (Document): The source document.
        offset (int): The character offset inside the document.

    Returns:
        int: The resulting document line.
    """
    line = max(document.source.count(NEW_LINE, 0, offset), 0)
    return line


def convert_document_offset_to_position(document: Document, offset: int) -> Position:
    """Converts the given offset in the document to a line/character based Position.

    Args:
        document (Document): The source document.
        offset (int): The character offset inside the document.

    Returns:
        Position: The resulting Position with line and character offset.
    """
    line = convert_document_offset_to_line(document, offset)
    line_offset = max(document.source.rfind(NEW_LINE, 0, offset), 0)
    character = offset - line_offset
    if line > 0:
        character -= 1
    return Position(line=line, character=character)


def convert_document_offsets_to_range(document: Document, start_offset: int, end_offset: int) -> Range:
    """Converts the given start and end offset positions in the document to a
    position Range based on line numbers.

    Args:
        start_offset (int): The start offset of the range
        end_offset (int): The end offset of the range

    Returns:
        Range: The resulting Range with the correct line number and character offset
    """
    return Range(
        start=convert_document_offset_to_position(document, start_offset),
        end=convert_document_offset_to_position(document, end_offset),
    )


class MultiLineStream:
    """Represents a multi-line stream of characters."""

    def __init__(self, source: str, position: int = 0) -> None:
        self._source = source
        self._position = position
        self._len = len(source)

    def eos(self) -> bool:
        """Indicates that the stream is at the end position."""
        return self._len <= self._position

    def get_source(self) -> str:
        """Gets the source text."""
        return self._source

    def pos(self) -> int:
        """Gets the current stream position."""
        return self._position

    def advance(self, n: int) -> None:
        """Advances the stream given number of positions."""
        self._position += n

    def go_to_end(self) -> None:
        """Sets the stream position to the end of the stream."""
        self._position = self._len

    def peek_char(self, n: int = 0) -> int:
        """Gets the value of the next position in the stream without advancing it."""
        try:
            return ord(self._source[self._position + n])
        except IndexError:
            return -1

    def advance_if_char(self, ch: int) -> bool:
        """If the next character in the stream matches the given character, the stream advances one position."""
        if ch == self.peek_char():
            self._position += 1
            return True
        return False

    def advance_if_chars(self, ch: List[int]) -> bool:
        """If the next characters in the stream matches the given sequence of characters, the stream advances
        the length of the sequence."""
        ch_length = len(ch)
        if self._position + ch_length > self._len:
            return False
        i = 0
        for i in range(ch_length):
            if self._source[self._position + i] != chr(ch[i]):
                return False
        self.advance(i)
        return True

    def advance_until_char(self, ch: int) -> bool:
        """Advances the stream until it founds a character matching the given."""
        while self._position < self._len:
            if ord(self._source[self._position]) == ch:
                return True
            self.advance(1)
        return False

    def advance_until_chars(self, ch: List[int]) -> bool:
        """Advances the stream until it founds a list of character matching the given."""
        ch_length = len(ch)
        while self._position + ch_length <= self._len:
            i = 0
            while i < ch_length and self._source[self._position + i] == chr(ch[i]):
                i += 1
            if i == ch_length:
                return True
            self.advance(i or 1)
        self.go_to_end()
        return False

    def advance_while_char_in(self, list: List[int]) -> int:
        """Advances the stream if the characters are like any of the characters in the given list."""
        pos_now = self._position
        while self._position < self._len and ord(self._source[self._position]) in list:
            self._position += 1
        return self._position - pos_now

    def advance_until_char_or_new_tag(self, ch: int) -> bool:
        """Advances the stream until it finds the given character or the '<' (new tag character)."""
        while self._position < self._len:
            if self.peek_char() in [ch, _LAN]:
                return True
            self.advance(1)
        return False

    def advance_until_chars_or_new_tag(self, ch: List[int]) -> bool:
        """Advances the stream until it finds the given sequence of characters or the '<' (new tag character)."""
        ch_length = len(ch)
        while self._position + ch_length <= self._len:
            i = 0
            if self.peek_char() == _LAN:
                return True
            while i < ch_length and self._source[self._position + i] == chr(ch[i]):
                i += 1
            if i == ch_length:
                return True
            self.advance(i or 1)
        self.go_to_end()
        return False

    def advance_while_char(self, predicate: Callable[[str], bool]) -> int:
        """Advances the stream while the given condition is True."""
        pos_now = self._position
        while self._position < self._len and predicate(self._source[self._position]):
            self._position += 1
        return self._position - pos_now

    def skip_whitespace(self) -> bool:
        """Advances the stream while any kind of white space character is found."""
        n = self.advance_while_char_in(WHITESPACE_CHARS)
        return n > 0



from typing import cast

from galaxy.util import xml_macros
from galaxyls.types import GeneratedExpandedDocument
from lxml import etree


def remove_macros(xml_tree: etree._ElementTree) -> etree._ElementTree:
    """Removes the macros section from the tool tree.

    Args:
        xml_tree (etree._ElementTree): The tool element tree.

    Returns:
        etree.ElementTree: The tool element tree without the macros section.
    """
    to_remove = []
    for macros_el in xml_tree.getroot().findall("macros"):
        to_remove.append(macros_el)
    for macros_el in to_remove:
        xml_tree.getroot().remove(macros_el)
    return xml_tree


class MacroExpanderService:
    def generate_expanded_from(self, tool_path: str) -> GeneratedExpandedDocument:
        result = GeneratedExpandedDocument()
        try:
            expanded_tool_tree, _ = xml_macros.load_with_references(tool_path)
            expanded_xml = remove_macros(expanded_tool_tree)
            root = expanded_xml.getroot()
            etree.indent(root, space=" " * 4)
            content = cast(str, etree.tostring(root, pretty_print=True, encoding=str))
            result.content = content
        except BaseException as e:
            result.error_message = f"{e}"
        return result



"""This module contains constants related to XSD processing.
"""

import galaxy.tool_util
from pathlib import Path

TOOL_XSD_FILE = Path(galaxy.tool_util.__file__).parent / "xsd" / "galaxy.xsd"

# Maximum depth when the schema contains recursive definitions
MAX_RECURSION_DEPTH = 25

# Schema namespace
XS_NAMESPACE = "{http://www.w3.org/2001/XMLSchema}"

# XSD elements
XS_ATTRIBUTE = f"{XS_NAMESPACE}attribute"
XS_COMPLEX_TYPE = f"{XS_NAMESPACE}complexType"
XS_ELEMENT = f"{XS_NAMESPACE}element"
XS_SCHEMA = f"{XS_NAMESPACE}schema"
XS_SEQUENCE = f"{XS_NAMESPACE}sequence"
XS_ALL = f"{XS_NAMESPACE}all"
XS_CHOICE = f"{XS_NAMESPACE}choice"
XS_SIMPLE_TYPE = f"{XS_NAMESPACE}simpleType"
XS_GROUP = f"{XS_NAMESPACE}group"
XS_ATTRIBUTE_GROUP = f"{XS_NAMESPACE}attributeGroup"
XS_SIMPLE_CONTENT = f"{XS_NAMESPACE}simpleContent"
XS_COMPLEX_CONTENT = f"{XS_NAMESPACE}complexContent"
XS_EXTENSION = f"{XS_NAMESPACE}extension"

# Messages
MSG_NO_DOCUMENTATION_AVAILABLE = "No documentation available"



"""This module contains definitions for constants used by the Galaxy Tools Language Server."""

SERVER_NAME = "Galaxy Tools LS"


class Commands:
    AUTO_CLOSE_TAGS = "galaxytools.completion.autoCloseTags"
    GENERATE_TESTS = "galaxytools.generate.tests"
    GENERATE_COMMAND = "galaxytools.generate.command"
    SORT_SINGLE_PARAM_ATTRS = "galaxytools.sort.singleParamAttributes"
    SORT_DOCUMENT_PARAMS_ATTRS = "galaxytools.sort.documentParamsAttributes"
    DISCOVER_TESTS = "galaxytools.tests.discover"
    GENERATE_EXPANDED_DOCUMENT = "galaxytools.generate.expandedDocument"


class DiagnosticCodes:
    INVALID_EXPANDED_TOOL = 101



"""Utilities to validate Galaxy xml tool wrappers and extract
information from the XSD schema.
"""

from typing import List

from lxml import etree
from pygls.lsp.types import Diagnostic, MarkupContent, MarkupKind

from ..context import XmlContext
from ..xml.document import XmlDocument
from .constants import MSG_NO_DOCUMENTATION_AVAILABLE, TOOL_XSD_FILE
from .parser import GalaxyToolXsdParser
from .validation import GalaxyToolValidationService

NO_DOC_MARKUP = MarkupContent(kind=MarkupKind.Markdown, value=MSG_NO_DOCUMENTATION_AVAILABLE)


class GalaxyToolXsdService:
    """Galaxy tool Xml Schema Definition service.

    This service provides functionality to extract information from
    the XSD schema and validate XML files against it.
    """

    def __init__(self, server_name: str) -> None:
        """Initializes the validator by loading the XSD."""
        self.server_name = server_name
        self.xsd_doc: etree._ElementTree = etree.parse(str(TOOL_XSD_FILE))
        self.xsd_schema = etree.XMLSchema(self.xsd_doc)
        self.xsd_parser = GalaxyToolXsdParser(self.xsd_doc.getroot())
        self.validator = GalaxyToolValidationService(server_name, self.xsd_schema)

    def validate_document(self, xml_document: XmlDocument) -> List[Diagnostic]:
        """Validates the Galaxy tool xml using the XSD schema and returns a list
        of diagnostics if there are any problems.
        """
        return self.validator.validate_document(xml_document)

    def get_documentation_for(self, context: XmlContext) -> MarkupContent:
        """Gets the documentation annotated in the XSD about the
        given element name (node or attribute).
        """
        if context.xsd_element:
            element = None
            if context.is_tag:
                element = context.xsd_element
            if context.is_attribute_key and context.node.name:
                element = context.xsd_element.attributes.get(context.node.name)
            if element:
                return element.get_doc()
        return NO_DOC_MARKUP



from typing import List, Optional

from galaxyls.services.definitions import DocumentDefinitionsProvider
from galaxyls.services.macros import MacroExpanderService
from galaxyls.services.tools.common import TestsDiscoveryService, ToolParamAttributeSorter
from galaxyls.services.tools.document import GalaxyToolXmlDocument
from galaxyls.services.tools.generators.command import GalaxyToolCommandSnippetGenerator
from galaxyls.services.tools.generators.tests import GalaxyToolTestSnippetGenerator
from galaxyls.services.tools.iuc import IUCToolParamAttributeSorter
from pygls.lsp.types import (
    CompletionList,
    CompletionParams,
    Diagnostic,
    DocumentFormattingParams,
    Hover,
    MarkupContent,
    MarkupKind,
    Position,
    Range,
    TextDocumentPositionParams,
    TextEdit,
)
from pygls.workspace import Document, Workspace
from galaxyls.services.tools.macros import MacroDefinitionsProvider
from galaxyls.services.tools.testing import ToolTestsDiscoveryService

from ..config import CompletionMode
from ..types import GeneratedSnippetResult, ReplaceTextRangeResult, TestSuiteInfoResult
from .completion import AutoCloseTagResult, XmlCompletionService
from .context import XmlContextService
from .format import GalaxyToolFormatService
from .xml.document import XmlDocument
from .xsd.service import GalaxyToolXsdService


class GalaxyToolLanguageService:
    """Galaxy tool Language service.

    This service manages all the operations supported
    by the LSP.
    """

    def __init__(self, server_name: str):
        self.xsd_service = GalaxyToolXsdService(server_name)
        self.format_service = GalaxyToolFormatService()
        self.xsd_tree = self.xsd_service.xsd_parser.get_tree()
        self.xml_context_service = XmlContextService(self.xsd_tree)
        self.sort_service: ToolParamAttributeSorter = IUCToolParamAttributeSorter()
        self.test_discovery_service: TestsDiscoveryService = ToolTestsDiscoveryService()
        self.macro_expander = MacroExpanderService()

    def set_workspace(self, workspace: Workspace):
        self.definitions_provider = DocumentDefinitionsProvider(MacroDefinitionsProvider(workspace))
        self.completion_service = XmlCompletionService(self.xsd_tree, self.definitions_provider)

    def get_diagnostics(self, xml_document: XmlDocument) -> List[Diagnostic]:
        """Validates the Galaxy tool XML document and returns a list
        of diagnostics if there are any problems.
        """
        return self.xsd_service.validate_document(xml_document)

    def get_documentation(self, xml_document: XmlDocument, position: Position) -> Optional[Hover]:
        """Gets the documentation about the element at the given position."""
        context = self.xml_context_service.get_xml_context(xml_document, position)
        if context.node:
            if context.is_tag or context.is_attribute_key:
                documentation = self.xsd_service.get_documentation_for(context)
                context_range = self.xml_context_service.get_range_for_context(xml_document, context)
                return Hover(contents=documentation, range=context_range)
            # Try to get token
            word = xml_document.document.word_at_position(position)
            token = self.definitions_provider.get_token_definition(xml_document, word)
            if token:
                return Hover(
                    contents=MarkupContent(kind=MarkupKind.Markdown, value=token.value),
                    range=Range(start=position, end=position),
                )
        return None

    def format_document(self, content: str, params: DocumentFormattingParams) -> List[TextEdit]:
        """Given the document contents returns the list of TextEdits
        needed to properly format and layout the document.
        """
        return self.format_service.format(content, params)

    def get_completion(
        self, xml_document: XmlDocument, params: CompletionParams, mode: CompletionMode
    ) -> Optional[CompletionList]:
        """Gets completion items depending on the current document context."""
        context = self.xml_context_service.get_xml_context(xml_document, params.position)
        return self.completion_service.get_completion_at_context(context, params.context, mode)

    def get_auto_close_tag(
        self, xml_document: XmlDocument, params: TextDocumentPositionParams
    ) -> Optional[AutoCloseTagResult]:
        """Gets the closing result for the currently opened tag in context."""
        # The trigger character `/` or `>` is placed right before the actual position, so we get the position.character - 1
        trigger_character = xml_document.document.lines[params.position.line][params.position.character - 1]
        # We want to get the context information right before the trigger character so we get position.character - 2
        position_before_trigger = Position(line=params.position.line, character=params.position.character - 2)
        context = self.xml_context_service.get_xml_context(xml_document, position_before_trigger)
        return self.completion_service.get_auto_close_tag(context, trigger_character)

    def generate_tests(self, document: Document) -> Optional[GeneratedSnippetResult]:
        """Generates a code snippet with some tests for the current inputs and outputs
        of this tool wrapper."""
        tool = GalaxyToolXmlDocument(document)
        generator = GalaxyToolTestSnippetGenerator(tool)
        return generator.generate_snippet()

    def generate_command(self, document: Document) -> Optional[GeneratedSnippetResult]:
        """Generates a boilerplate Cheetah code snippet based on the current inputs and outputs
        of this tool wrapper."""
        tool = GalaxyToolXmlDocument(document)
        generator = GalaxyToolCommandSnippetGenerator(tool)
        return generator.generate_snippet()

    def sort_single_param_attrs(
        self, xml_document: XmlDocument, params: TextDocumentPositionParams
    ) -> Optional[ReplaceTextRangeResult]:
        """Sorts the attributes of the param element under the cursor."""
        offset = xml_document.document.offset_at_position(params.position)
        param_element = xml_document.find_element_at(offset)
        if param_element:
            return self.sort_service.sort_param_attributes(param_element, xml_document)
        return None

    def sort_document_param_attributes(self, xml_document: XmlDocument) -> List[ReplaceTextRangeResult]:
        """Sorts the attributes of all the param elements contained in the document."""
        return self.sort_service.sort_document_param_attributes(xml_document)

    def discover_tests(self, workspace: Workspace) -> List[TestSuiteInfoResult]:
        """Sorts the attributes of all the param elements contained in the document."""
        return self.test_discovery_service.discover_tests_in_workspace(workspace)



from abc import ABC, abstractmethod
from typing import Dict, List, Optional, Tuple, cast

from anytree import NodeMixin
from galaxyls.services.xml.constants import UNDEFINED_OFFSET
from galaxyls.services.xml.types import NodeType


class XmlSyntaxNode(ABC, NodeMixin):
    """Abstract base class that represents a syntax node in the syntax tree."""

    def __init__(self):
        self.name: Optional[str] = None
        self.start: int = UNDEFINED_OFFSET
        self.end: int = UNDEFINED_OFFSET
        self._closed: bool = False

    @property
    def is_closed(self) -> bool:
        """Indicates if this node has been closed."""
        return self._closed

    @property
    def is_element(self) -> bool:
        """Indicates if this node is an element node."""
        return self.node_type == NodeType.ELEMENT

    @property
    def is_attribute(self) -> bool:
        """Indicates if this node is an attribute node."""
        return self.node_type in [NodeType.ATTRIBUTE, NodeType.ATTRIBUTE_KEY, NodeType.ATTRIBUTE_VALUE]

    @property
    def has_attributes(self) -> bool:
        """Indicates if this node has any attributes defined."""
        return False

    @property
    def node_type(self) -> NodeType:
        """The type of this node."""
        return NodeType.UNKNOWN

    @property
    def stack(self) -> List[str]:
        """The list of names of the ancestor elements including the actual element."""
        stack = [element.name for element in self.ancestors if element.name and type(element) is XmlElement]
        if self.is_element and self.name:
            stack.append(self.name)
        return stack

    def is_at(self, offset: int) -> bool:
        """Indicates if the offset is within this node definition."""
        return self.start <= offset < self.end

    def is_at_closing_tag(self, offset: int) -> bool:
        """Indicates if the offset is within an element's closing tag."""
        return False

    def get_attribute_nodes(self) -> List["XmlAttribute"]:
        """Gets the lists of attributes of this node if it has any."""
        return []

    def get_attribute_names(self) -> List[str]:
        """Gets the list of attribute names of this node if it has any."""
        return []

    def get_attribute_name(self) -> Optional[str]:
        """Gets the name of this attribute (if it is an attribute node)."""
        return None

    def find_node_at(self, offset: int) -> Optional["XmlSyntaxNode"]:
        """Finds the syntax node at the given document offset."""
        try:
            child = next(child for child in self.children if child.is_at(offset))
            return child.find_node_at(offset)
        except StopIteration:
            if self.is_at(offset):
                return self.find_attr_node_at(offset)
            return self

    def find_attr_node_at(self, offset: int) -> Optional["XmlSyntaxNode"]:
        """Finds the attribute node at the given document offset."""
        if self.has_attributes:
            attr_nodes = self.get_attribute_nodes()
            for attr in attr_nodes:
                if attr.is_at(offset):
                    return attr
        return self

    def find_element_at(self, offset: int) -> Optional["XmlElement"]:
        """Finds the element node at the given document offset."""
        node = self.find_node_at(offset)
        if node:
            return node.get_parent_element()
        return None

    def get_parent_element(self) -> Optional["XmlElement"]:
        """Returns the XmlElement associated with this syntax node or None."""
        if self.is_element:
            return cast(XmlElement, self)
        if self.parent:
            parent: Optional[XmlSyntaxNode] = cast(XmlSyntaxNode, self.parent)
            return parent.get_parent_element()
        return None

    def get_offsets(self, _: int) -> Tuple[int, int]:
        """Get the starting and ending offsets of this syntax node as a tuple.

        Args:
            offset (int): The current document offset. Used to determine if the
            offset is currently over the starting or closing tag.

        Returns:
            Tuple[int, int]: The start and end offsets of this tag.
        """
        if self.name:
            return self.start, self.start + len(self.name)
        return self.start, self.end


class XmlContainerNode(XmlSyntaxNode):
    """Represents a node that can have content."""

    @abstractmethod
    def get_content_offsets(self) -> Tuple[int, int]:
        return NotImplemented

    def get_content(self, source: str) -> Optional[str]:
        """Gets the text content from the source between the content offsets of this element."""
        start, end = self.get_content_offsets()
        return source[start:end]


class XmlContent(XmlContainerNode):
    """Represents some content inside a XML document."""

    def __init__(self, start: int, end: int):
        super().__init__()
        self.start = start
        self.end = end

    @property
    def node_type(self) -> NodeType:
        """The type of this node."""
        return NodeType.CONTENT

    def get_content_offsets(self) -> Tuple[int, int]:
        return self.start, self.end

    def is_at(self, offset: int) -> bool:
        """Indicates if the offset is within this node definition."""
        return self.start <= offset <= self.end


class XmlAttribute(XmlSyntaxNode):
    """Represents an attribute of a XML element."""

    def __init__(self, name: str, start: int, end: int, owner: "XmlElement"):
        super().__init__()
        self.name = name
        self.start = start
        self.end = end
        self.owner = owner
        self.key = XmlAttributeKey(name, start, end, self)
        self.key.parent = self
        self.has_delimiter: bool = False
        self.value: Optional[XmlAttributeValue] = None
        self.parent = owner

    def __repr__(self) -> str:
        return f"[{self.name}={self.get_value()}]"

    @property
    def node_type(self) -> NodeType:
        """The type of this node."""
        return NodeType.ATTRIBUTE

    def set_value(self, value: Optional[str], start: int, end: int) -> None:
        """Sets the value of this attribute."""
        if self.value:
            self.value.update(value, start, end)
        else:
            self.value = XmlAttributeValue(value, start, end, self)
        self.end = end

    def get_value(self) -> str:
        """Gets the value of this attribute (unquoted) or an empty str."""
        try:
            return self.value.unquoted
        except BaseException:
            return ""

    def get_attribute_nodes(self) -> List["XmlAttribute"]:
        """Gets the lists of attributes of this node if it has any."""
        return cast(List[XmlAttribute], self.children)

    def get_attribute_name(self) -> Optional[str]:
        """Gets the name of this attribute (if it is an attribute node)."""
        return self.name


class XmlAttributeKey(XmlSyntaxNode):
    """Represents the key (name) of a XML attribute."""

    def __init__(self, name: str, start: int, end: int, owner: XmlAttribute):
        super().__init__()
        self.name = name
        self.start = start
        self.end = end
        self.owner = owner

    @property
    def node_type(self) -> NodeType:
        """The type of this node."""
        return NodeType.ATTRIBUTE_KEY

    def get_attribute_name(self) -> Optional[str]:
        """Gets the name of this attribute (if it is an attribute node)."""
        return self.name

    def get_offsets(self, _: int) -> Tuple[int, int]:
        """Get the starting and ending offsets of this syntax node as a tuple.

        Args:
            offset (int): Not used in this override.

        Returns:
            Tuple[int, int]: The start and end offsets of this tag.
        """
        return self.start, self.end


class XmlAttributeValue(XmlContainerNode):
    """Represents the value of a XML attribute."""

    def __init__(self, value: Optional[str], start: int, end: int, owner: XmlAttribute):
        super().__init__()
        self.quoted = value
        self.start = start
        self.end = end
        self.owner = owner
        self.parent = owner

    def update(self, value: Optional[str], start: int, end: int) -> None:
        self.quoted = value
        self.start = start
        self.end = end

    @property
    def node_type(self) -> NodeType:
        """The type of this node."""
        return NodeType.ATTRIBUTE_VALUE

    @property
    def unquoted(self) -> str:
        """The value without quoting marks."""
        return self.quoted.strip("\"'")

    def get_attribute_name(self) -> Optional[str]:
        """Gets the name of this attribute (if it is an attribute node)."""
        return self.owner.name

    def get_content_offsets(self) -> Tuple[int, int]:
        return self.start, self.end


class XmlElement(XmlContainerNode):
    """Represents a XML element in the document syntax tree."""

    def __init__(self, start: int = UNDEFINED_OFFSET, end: int = UNDEFINED_OFFSET):
        super().__init__()
        self.name: Optional[str] = None
        self.start = start
        self.end = end
        self.start_tag_open_offset: int = UNDEFINED_OFFSET
        self.start_tag_close_offset: int = UNDEFINED_OFFSET
        self.end_tag_open_offset: int = UNDEFINED_OFFSET
        self.end_tag_close_offset: int = UNDEFINED_OFFSET
        self.is_self_closed: bool = False
        self.attributes: Dict[str, XmlAttribute] = {}

    def __repr__(self) -> str:
        return f"XmlElement[{self.name}]-Attrs[{','.join(self.get_attribute_names())}]"

    @property
    def node_type(self) -> NodeType:
        """The type of this node."""
        return NodeType.ELEMENT

    @property
    def has_attributes(self) -> bool:
        """Indicates if this node has any attributes defined."""
        return len(self.attributes) > 0

    @property
    def elements(self) -> List["XmlElement"]:
        """The child elements of this element."""
        return [element for element in self.children if type(element) is XmlElement]

    @property
    def name_start_offset(self) -> int:
        """The offset at where this element's tag name starts."""
        return self.start + 1  # +1 for '<'

    @property
    def name_end_offset(self) -> int:
        """The offset at where this element's tag name ends."""
        if self.name:
            return self.name_start_offset + len(self.name)
        return self.name_start_offset

    @property
    def end_offset(self) -> int:
        """The offset position at the end of this element"""
        if self.end_tag_close_offset != UNDEFINED_OFFSET:
            return self.end_tag_close_offset + 1  # +1 for '>'
        if self.start_tag_close_offset != UNDEFINED_OFFSET:
            return self.start_tag_close_offset + 1  # +1 for '>'
        return UNDEFINED_OFFSET

    def is_same_tag(self, tag: str) -> bool:
        """Indicates if this element has the same tag name as the one provided."""
        return self.name == tag

    def is_at(self, offset: int) -> bool:
        """Indicates if the offset is within this node definition."""
        if self.end_tag_close_offset != UNDEFINED_OFFSET:
            return self.start <= offset <= self.end_tag_close_offset
        return self.start <= offset <= self.end

    def is_at_closing_tag(self, offset: int) -> bool:
        """Indicates if the offset is within an element's closing tag."""
        return self.end_tag_open_offset <= offset <= self.end_tag_close_offset

    def get_attribute_nodes(self) -> List[XmlAttribute]:
        """Gets the lists of attributes of this node if it has any."""
        result: List[XmlAttribute] = []
        for attr in self.attributes.values():
            result.extend(attr.get_attribute_nodes())
        return result

    def get_attribute_names(self) -> List[str]:
        """Gets the list of attribute names of this node if it has any."""
        return [*self.attributes]

    def get_attribute(self, name: str) -> Optional[str]:
        """Gets the value of the attribute with the given name if it exists.

        Args:
            name (str): The name of the attribute which value will be returned.

        Returns:
            Optional[str]: The value of the attribute or None if it does not exists.
        """
        try:
            return self.attributes[name].get_value()
        except BaseException:
            return None

    def get_offsets(self, offset: int) -> Tuple[int, int]:
        """Get the starting and ending offsets of this element's name as a tuple.

        Args:
            offset (int): The current document offset. Used to determine if the
            offset is currently over the starting or closing tag.

        Returns:
            Tuple[int, int]: The start and end offsets of this tag name whitout the
            opening and ending tokens ('<', '>' '</').
        """
        if self.name:
            start = self.name_start_offset
            end = self.name_end_offset
            if self.is_at_closing_tag(offset):
                start = self.end_tag_open_offset + 2  # +1 for '</'
                end = self.end_tag_close_offset
            return start, end
        return self.start, self.end

    def get_attributes_offsets(self) -> Tuple[int, int]:
        """Get the starting and ending offsets of all the attributes of this element.

        Returns:
            Tuple[int, int]: The start and end offsets of the attributes. The starting offset
            will be the start of the first attribute and the end offset the end of the last
            attribute defined in this element.
        """
        if self.has_attributes:
            attributes = self.get_attribute_nodes()
            first_attr = attributes[0]
            last_attr = attributes[-1]
            start = first_attr.start
            end = last_attr.end
            return start, end

        if self.name:
            return self.name_end_offset, self.name_end_offset

        return self.end, self.end

    def get_content_offsets(self) -> Tuple[int, int]:
        start = self.start_tag_close_offset + 1  # +1 for '>'
        end = self.end_tag_open_offset
        if self.is_self_closed:
            start = end = -1
        return start, end

    def get_children_with_name(self, name: str) -> List["XmlElement"]:
        children = [child for child in self.children if child.name == name]
        return list(children)

    def get_cdata_section(self) -> Optional["XmlCDATASection"]:
        """Gets the CDATA node inside this element or None if it doesn't have a CDATA section."""
        return next((node for node in self.children if type(node) == XmlCDATASection), None)


class XmlCDATASection(XmlContainerNode):
    """Represents a CDATA section in a XML document."""

    def __init__(self, start: int, end: int):
        super().__init__()
        self.start = start
        self.end = end
        self.start_content: int = UNDEFINED_OFFSET
        self.end_content: int = UNDEFINED_OFFSET

    @property
    def node_type(self) -> NodeType:
        """The type of this node."""
        return NodeType.CDATA_SECTION

    def get_content_offsets(self) -> Tuple[int, int]:
        return self.start_content, self.end_content


class XmlComment(XmlSyntaxNode):
    """Represents a comment section in a XML document."""

    def __init__(self, start: int, end: int):
        super().__init__()
        self.start = start
        self.end = end
        self.start_content: int = UNDEFINED_OFFSET
        self.end_content: int = UNDEFINED_OFFSET

    @property
    def node_type(self) -> NodeType:
        """The type of this node."""
        return NodeType.COMMENT


class XmlProcessingInstruction(XmlSyntaxNode):
    """Represents a processing instruction (like the prolog) in a XML document."""

    def __init__(self, start: int, end: int):
        super().__init__()
        self.start = start
        self.end = end
        self.start_content: int = UNDEFINED_OFFSET
        self.end_content: int = UNDEFINED_OFFSET

    @property
    def node_type(self) -> NodeType:
        """The type of this node."""
        return NodeType.PROCESSING_INSTRUCTION



from typing import List, Optional, cast

from pygls.lsp.types import Location, Position
from galaxyls.services.tools.macros import MacroDefinitionsProvider, TokenDefinition
from galaxyls.services.xml.document import XmlDocument
from galaxyls.services.xml.nodes import XmlContent


class DocumentDefinitionsProvider:
    """Provides symbol definition locations."""

    def __init__(self, macro_definitions_provider: MacroDefinitionsProvider) -> None:
        self.macro_definitions_provider = macro_definitions_provider

    def go_to_definition(self, xml_document: XmlDocument, position: Position) -> Optional[List[Location]]:
        macro_definitions = self.macro_definitions_provider.load_macro_definitions(xml_document)
        word = xml_document.document.word_at_position(position)
        token_definition = macro_definitions.get_token_definition(word)
        if token_definition:
            return [token_definition.location]

        macro_definition = macro_definitions.get_macro_definition(word)
        if macro_definition:
            return [macro_definition.location]

        offset = xml_document.document.offset_at_position(position)
        node = xml_document.find_node_at(offset)
        if isinstance(node, XmlContent) and node.parent.name == "import":
            content_node = cast(XmlContent, node)
            start, end = content_node.get_content_offsets()
            import_filename = xml_document.get_text_between_offsets(start, end)
            return macro_definitions.go_to_import_definition(import_filename)

    def get_token_definition(self, xml_document: XmlDocument, token: str) -> Optional[TokenDefinition]:
        macro_definitions = self.macro_definitions_provider.load_macro_definitions(xml_document)
        return macro_definitions.get_token_definition(token)



from typing import Dict, List, Optional, Set

from pydantic import BaseModel
from pygls.lsp.types import Location
from pygls.workspace import Workspace

from galaxyls.services.tools.constants import MACRO, NAME, TOKEN, XML
from galaxyls.services.tools.document import GalaxyToolXmlDocument
from galaxyls.services.xml.document import XmlDocument
from galaxyls.services.xml.nodes import XmlElement
from galaxyls.services.xml.parser import XmlDocumentParser


class BaseMacrosModel(BaseModel):
    class Config:
        arbitrary_types_allowed = True


class TokenDefinition(BaseModel):
    name: str
    location: Location
    value: str


class TokenParam(TokenDefinition):
    param_name: str
    default_value: str


class MacroDefinition(BaseModel):
    name: str
    location: Location
    token_params: Dict[str, TokenParam]


class ImportedMacrosFile(BaseMacrosModel):
    file_name: str
    file_uri: Optional[str]
    document: Optional[XmlDocument]
    tokens: Dict[str, TokenDefinition]
    macros: Dict[str, MacroDefinition]


class ToolMacroDefinitions(BaseMacrosModel):
    tool_document: XmlDocument
    imported_macros: Dict[str, ImportedMacrosFile]
    tokens: Dict[str, TokenDefinition]
    macros: Dict[str, MacroDefinition]

    def go_to_import_definition(self, file_name: str) -> Optional[List[Location]]:
        imported_macros = self.imported_macros.get(file_name)
        if imported_macros and imported_macros.document and imported_macros.document.root:
            macros_file_uri = imported_macros.file_uri
            content_range = imported_macros.document.get_full_range(imported_macros.document.root)
            if content_range:
                return [
                    Location(
                        uri=macros_file_uri,
                        range=content_range,
                    )
                ]

    def get_token_definition(self, token: str) -> Optional[TokenDefinition]:
        definition = self.tokens.get(token)
        if definition is None:
            return self.get_token_param_definition(token)
        return definition

    def get_token_param_definition(self, token: str) -> Optional[TokenParam]:
        for macro in self.macros.values():
            if token in macro.token_params:
                return macro.token_params.get(token)
        return None

    def get_macro_definition(self, macro_name: str) -> Optional[MacroDefinition]:
        return self.macros.get(macro_name)


class MacroDefinitionsProvider:
    """Provides location information about macros imported by a tool."""

    def __init__(self, workspace: Workspace) -> None:
        self.workspace = workspace

    def load_macro_definitions(self, tool_xml: XmlDocument) -> ToolMacroDefinitions:
        tool = GalaxyToolXmlDocument.from_xml_document(tool_xml)
        tokens = self._get_token_definitions(tool_xml)
        macros = self._get_macro_definitions(tool_xml)
        imported_macro_files = self._get_imported_macro_files_from_tool(tool)
        for file in imported_macro_files.values():
            tokens.update(file.tokens)
            macros.update(file.macros)
        return ToolMacroDefinitions(
            tool_document=tool_xml,
            imported_macros=imported_macro_files,
            tokens=tokens,
            macros=macros,
        )

    def get_macro_names(self, tool_xml: XmlDocument) -> Set[str]:
        tool = GalaxyToolXmlDocument.from_xml_document(tool_xml)
        macros = self._get_macro_definitions(tool_xml)
        imported_macro_files = self._get_imported_macro_files_from_tool(tool)
        for file in imported_macro_files.values():
            macros.update(file.macros)
        return set(macros.keys())

    def get_macro_token_params(self, tool_xml: XmlDocument, macro_name: str) -> List[TokenParam]:
        macro_definitions = self.load_macro_definitions(tool_xml)
        macro = macro_definitions.get_macro_definition(macro_name)
        if macro.token_params:
            return list(macro.token_params.values())
        return []

    def _get_imported_macro_files_from_tool(self, tool: GalaxyToolXmlDocument) -> Dict[str, ImportedMacrosFile]:
        macro_files = {}
        uris_dict = tool.get_macro_import_uris()
        for file_name, file_uri in uris_dict.items():
            macros_document = self._load_macros_document(file_uri)
            tokens = self._get_token_definitions(macros_document)
            macros = self._get_macro_definitions(macros_document)
            macro_files[file_name] = ImportedMacrosFile(
                file_name=file_name,
                file_uri=file_uri,
                document=macros_document,
                tokens=tokens,
                macros=macros,
            )
        return macro_files

    def _load_macros_document(self, document_uri: str) -> XmlDocument:
        document = self.workspace.get_document(document_uri)
        xml_document = XmlDocumentParser().parse(document)
        return xml_document

    def _get_token_definitions(self, macros_xml: XmlDocument) -> Dict[str, TokenDefinition]:
        token_elements = macros_xml.find_all_elements_with_name(TOKEN)
        rval = {}
        for element in token_elements:
            token = element.get_attribute(NAME)
            name = token.replace("@", "")
            value = element.get_content(macros_xml.document.source)
            token_def = TokenDefinition(
                name=name,
                location=Location(
                    uri=macros_xml.document.uri,
                    range=macros_xml.get_full_range(element),
                ),
                value=value,
            )
            rval[token_def.name] = token_def
        return rval

    def _get_macro_definitions(self, macros_xml: XmlDocument) -> Dict[str, MacroDefinition]:
        macro_elements = macros_xml.find_all_elements_with_name(MACRO)
        xml_elements = macros_xml.find_all_elements_with_name(XML)
        macro_elements += xml_elements
        rval = {}
        for element in macro_elements:
            name = element.get_attribute(NAME)
            macro_def = MacroDefinition(
                name=name,
                location=Location(
                    uri=macros_xml.document.uri,
                    range=macros_xml.get_full_range(element),
                ),
                token_params=self.get_token_params(macros_xml, element),
            )
            rval[macro_def.name] = macro_def
        return rval

    def get_token_params(self, macros_xml: XmlDocument, element: XmlElement) -> Dict[str, TokenParam]:
        token_params = {}
        for attr_name, attr in element.attributes.items():
            if attr_name.startswith("token_"):
                param_name = attr_name.replace("token_", "")
                token_name = param_name.upper()
                default_value = attr.get_value()
                token_param = TokenParam(
                    name=token_name,
                    param_name=param_name,
                    default_value=default_value,
                    value=f"**Token parameter**\n- Default value: `{default_value}`",
                    location=Location(
                        uri=macros_xml.document.uri,
                        range=macros_xml.get_full_range(attr),
                    ),
                )
                token_params[token_name] = token_param
        return token_params



from pathlib import Path
from typing import Dict, List, Optional, cast

from anytree import find
from pygls.lsp.types import Position, Range
from pygls.workspace import Document
from galaxyls.services.tools.constants import IMPORT, INPUTS, MACROS, OUTPUTS, TESTS, TOOL
from galaxyls.services.tools.inputs import GalaxyToolInputTree
from galaxyls.services.xml.nodes import XmlContainerNode, XmlElement

from galaxyls.services.xml.types import DocumentType
from galaxyls.services.xml.document import XmlDocument
from galaxyls.services.xml.parser import XmlDocumentParser


class GalaxyToolXmlDocument:
    """Represents a Galaxy tool XML wrapper.

    This class provides access to the tool definitions and some utilities to extract
    information from the document.
    """

    def __init__(self, document: Document, xml_document: Optional[XmlDocument] = None) -> None:
        if xml_document:
            self.xml_document = xml_document
            self.document: Document = xml_document.document
        else:
            self.document: Document = document
            self.xml_document = XmlDocumentParser().parse(document)

    @property
    def is_valid(self) -> bool:
        """Indicates if this document is a valid Galaxy Tool Wrapper
        XML document."""
        return self.xml_document.document_type == DocumentType.TOOL

    @property
    def uses_macros(self) -> bool:
        """Indicates if this tool document *uses* macro definitions.

        Returns:
            bool: True if the tool contains <expand> elements.
        """
        return self.xml_document.uses_macros

    @property
    def source(self) -> str:
        """The contents of the tool document."""
        return self.xml_document.document.source

    @property
    def path(self) -> str:
        """The file path of the tool."""
        return self.xml_document.document.path

    def find_element(self, name: str, maxlevel: int = 3) -> Optional[XmlElement]:
        """Finds the element with the given name in the document.

        Args:
            name (str): The name of the element to find.
            maxlevel (int, optional): The level at which the search will
            stop if not found. Defaults to 3.

        Returns:
            Optional[XmlElement]: The first element matching the name.
        """
        node = find(self.xml_document, filter_=lambda node: node.name == name, maxlevel=maxlevel)
        return cast(XmlElement, node)

    def get_content_range(self, element: Optional[XmlContainerNode]) -> Optional[Range]:
        """Returns the Range of the content block of the given element.

        Args:
            element (Optional[XmlElement]): The element.

        Returns:
            Optional[Range]: The Range of the content block.
        """
        if element:
            return self.xml_document.get_content_range(element)

    def get_position_before(self, element: XmlElement) -> Position:
        """Returns the document position right before the given element opening tag.

        Args:
            element (XmlElement): The element.

        Returns:
            Position: The position right before the element opening tag.
        """
        return self.xml_document.get_position_before(element)

    def get_position_after(self, element: XmlElement) -> Position:
        """Returns the document position right after the given element closing tag.

        Args:
            element (XmlElement): The element.

        Returns:
            Position: The position right after the element closing tag.
        """
        return self.xml_document.get_position_after(element)

    def analyze_inputs(self) -> GalaxyToolInputTree:
        """Gets the inputs in the document and builds the input tree.

        Returns:
            GalaxyToolInputTree: The resulting input tree for this document.
        """
        inputs = self.find_element(INPUTS)
        return GalaxyToolInputTree(inputs)

    def get_outputs(self) -> List[XmlElement]:
        """Gets the outputs of this document as a list of elements.

        Returns:
            List[XmlElement]: The outputs defined in the document.
        """
        outputs = self.find_element(OUTPUTS)
        if outputs:
            return outputs.elements
        return []

    def get_tool_element(self) -> Optional[XmlElement]:
        """Gets the root tool element"""
        return self.find_element(TOOL)

    def get_macros_element(self) -> Optional[XmlElement]:
        """Gets the macros element"""
        return self.find_element(MACROS)

    def get_macro_import_elements(self) -> Optional[List[XmlElement]]:
        """Gets the list of elements with macro file import declarations."""
        return self.xml_document.find_all_elements_with_name(IMPORT)

    def get_macro_import_uris(self) -> Dict[str, str]:
        """Returns a dictionary with the file name and corresponding file uri of each
        imported macro file."""
        result = {}
        tool_directory = self._get_tool_directory()
        import_elements = self.get_macro_import_elements()
        for imp in import_elements:
            filename = imp.get_content(self.xml_document.document.source)
            if filename:
                path = tool_directory / filename
                if path.exists():
                    file_uri = path.as_uri()
                    result[filename] = file_uri
        return result

    def get_macros_range(self) -> Optional[Range]:
        """Returns the Range position of the macros element name if it exists."""
        element = self.get_macros_element()
        if element:
            range = self.xml_document.get_element_name_range(element)
            return range

    def get_import_macro_file_range(self, file_path: Optional[str]) -> Optional[Range]:
        """Returns the Range position of the imported macro file element if it exists."""
        if file_path:
            filename = Path(file_path).name
            import_elements = self.get_macro_import_elements()
            for imp in import_elements:
                imp_filename = imp.get_content(self.xml_document.document.source)
                if imp_filename == filename:
                    return self.xml_document.get_full_range(imp)

    def get_tool_id(self) -> Optional[str]:
        """Gets the identifier of the tool"""
        tool_element = self.get_tool_element()
        return tool_element.get_attribute("id")

    def get_tests(self) -> List[XmlElement]:
        """Gets the tests of this document as a list of elements.

        Returns:
            List[XmlElement]: The tests defined in the document.
        """
        tests = self.find_element(TESTS)
        if tests:
            return tests.elements
        return []

    @classmethod
    def from_xml_document(cls, xml_document: XmlDocument) -> "GalaxyToolXmlDocument":
        return GalaxyToolXmlDocument(xml_document.document, xml_document)

    def _get_tool_directory(self):
        tool_directory = Path(self.xml_document.document.path).resolve().parent
        return tool_directory



"""Module for converting XSD schema to tree structure
to simplify access to definitions.
"""

from lxml import etree
from typing import List, Optional, Dict, cast
from .types import XsdNode, XsdAttribute, XsdTree

from .constants import (
    MAX_RECURSION_DEPTH,
    XS_ALL,
    XS_ATTRIBUTE,
    XS_ATTRIBUTE_GROUP,
    XS_CHOICE,
    XS_COMPLEX_TYPE,
    XS_ELEMENT,
    XS_GROUP,
    XS_SEQUENCE,
    XS_SIMPLE_TYPE,
    XS_SIMPLE_CONTENT,
    XS_COMPLEX_CONTENT,
    XS_EXTENSION,
)


class GalaxyToolXsdParser:
    """Parser for converting the XSD schema into a tree structure
    that it is easy to use to retrieve information about the schema.

    Note: the conversion is focused in supporting the definitions
    used in the Galaxy tool's XSD and not all the possible definitions
    supported by the XSD specification.
    """

    def __init__(self, xsd_root: etree._Element):
        """Initializes the parser with the given etree root element.

        Args:
            xsd_root (etree._Element): The root element of the XSD.
        """
        self._root: etree._Element = xsd_root
        self._tree: Optional[XsdTree] = None
        self._named_type_map: Dict[str, etree._Element] = {}
        self._named_group_map: Dict[str, etree._Element] = {}

    def get_tree(self) -> XsdTree:
        """Builds the tree structure from the root etree._Element if
        it does not exists, or returns it if exists.

        Returns:
            XsdTree: The tree structure with all the relevant
            information from the XSD schema.
        """
        if self._tree is None:
            self._register_named_types()
            self._register_named_groups()
            self._build_tree_recursive(self._root, self._tree)
            if self._tree is None:
                self._tree = self._build_empty_tree()
        return cast(XsdTree, self._tree)

    def _build_empty_tree(self) -> XsdTree:
        return XsdTree(XsdNode(""))

    def _register_named_types(self) -> None:
        """Finds all (named) complex and simple type definitions in the
        XSD schema and register them in a dictionary.
        """
        for element in self._root.findall(XS_COMPLEX_TYPE):
            name = element.get("name")
            if name:
                self._named_type_map[name] = element

        for element in self._root.findall(XS_SIMPLE_TYPE):
            name = element.get("name")
            if name:
                self._named_type_map[name] = element

    def _register_named_groups(self) -> None:
        """Finds all (named) groups and attribute group definitions in the
        XSD schema and register them in a dictionary.
        """
        for element in self._root.findall(XS_GROUP):
            name = element.get("name")
            if name:
                self._named_group_map[name] = element

        for element in self._root.findall(XS_ATTRIBUTE_GROUP):
            name = element.get("name")
            if name:
                self._named_group_map[name] = element

    def _build_tree_recursive(self, parent_element: etree._Element, parent_node: Optional[XsdNode], depth: int = 0) -> None:
        if depth > MAX_RECURSION_DEPTH:
            return None  # Stop recursion

        for element in parent_element:
            tag = element.tag
            element_name = element.attrib.get("name")
            element_type_name = element.attrib.get("type")
            if tag == XS_ELEMENT:
                node = XsdNode(element_name, element, parent_node)
                if not parent_node:  # Is root element
                    self._tree = XsdTree(node)
                node.xsd_type = self._named_type_map.get(element_type_name)
                self._apply_named_type_to_node(element_type_name, node, depth + 1)
                # minOccurs defaults to 1
                node.min_occurs = int(element.attrib.get("minOccurs", 1))
                max_occurs = element.attrib.get("maxOccurs", -1)
                if max_occurs == "unbounded":
                    max_occurs = -1
                node.max_occurs = int(max_occurs)
                self._build_tree_recursive(element, node, depth + 1)
            elif tag == XS_COMPLEX_TYPE:
                if not element_name:
                    # The type is anonymous and is declared inside the element
                    # so we can directly apply it to the node
                    self._apply_complex_type_to_node(element, parent_node, depth + 1)
            elif tag == XS_GROUP:
                element_ref = element.attrib.get("ref")
                if element_ref:
                    self._apply_group_to_node(element_ref, parent_node, depth + 1)

    def _apply_named_type_to_node(self, type_name: str, node: XsdNode, depth: int = 0) -> None:
        element_type = self._named_type_map.get(type_name)
        if element_type is not None and element_type.tag == XS_COMPLEX_TYPE:
            self._apply_complex_type_to_node(element_type, node, depth + 1)

    def _apply_complex_type_to_node(self, complex_type: etree._Element, node: XsdNode, depth: int = 0) -> None:
        for child_element in complex_type:
            tag = child_element.tag
            if tag in [XS_ALL, XS_SEQUENCE, XS_CHOICE]:
                self._build_tree_recursive(child_element, node, depth + 1)
            elif tag == XS_ATTRIBUTE:
                self._add_attribute_to_node(child_element, node)
            elif tag == XS_ATTRIBUTE_GROUP:
                element_ref = child_element.attrib.get("ref")
                if element_ref:
                    self._apply_attribute_group_to_node(element_ref, node)
            elif tag == XS_SIMPLE_CONTENT:
                self._apply_simple_content_to_node(child_element, node)
            elif tag == XS_COMPLEX_CONTENT:
                self._apply_complex_content_to_node(child_element, node, depth)

    def _apply_attribute_group_to_node(self, group_name: str, node: XsdNode) -> None:
        group = self._named_group_map.get(group_name)
        if group is not None:
            for child_element in group:
                self._add_attribute_to_node(child_element, node)

    def _apply_group_to_node(
        self,
        group_name: str,
        node: XsdNode,
        depth: int,
    ) -> None:
        group = self._named_group_map.get(group_name)
        if group is not None:
            for child_element in group:
                self._build_tree_recursive(child_element, node, depth + 1)

    def _apply_simple_content_to_node(self, simple_content: etree._Element, node: XsdNode) -> None:
        for child_elem in simple_content:
            if child_elem.tag == XS_EXTENSION:
                attributes = child_elem.findall(XS_ATTRIBUTE)
                for attr_elem in attributes:
                    self._add_attribute_to_node(attr_elem, node)

    def _apply_complex_content_to_node(self, complex_content: etree._Element, node: XsdNode, depth: int) -> None:
        for child_elem in complex_content:
            if child_elem.tag == XS_EXTENSION:
                attr_base = child_elem.attrib.get("base")
                if attr_base:
                    self._apply_named_type_to_node(attr_base, node, depth)
                self._apply_complex_type_to_node(child_elem, node, depth)

    def _add_attribute_to_node(self, attribute_element: etree._Element, node: XsdNode) -> None:
        attr_name = attribute_element.attrib.get("name")
        attr_type = attribute_element.attrib.get("type")
        attr_use = attribute_element.attrib.get("use")
        attr = XsdAttribute(attr_name, attribute_element, attr_type, attr_use == "required")
        attr.enumeration = self._get_enumeration_restrictions_from_type(attr_type)
        node.attributes[attr_name] = attr

    def _get_enumeration_restrictions_from_type(self, type_name: str) -> Optional[List[str]]:
        simple_type = self._named_type_map.get(type_name)
        if simple_type is None:
            return []
        if simple_type.tag == XS_SIMPLE_TYPE:
            return simple_type.xpath(".//xs:enumeration/@value", namespaces=simple_type.nsmap)



""" This code is based on the Eclipse/Lemminx XML language server implementation:
https://github.com/eclipse/lemminx/tree/master/org.eclipse.lemminx/src/main/java/org/eclipse/lemminx/dom

Only the minimum subset of the XML dialect used by Galaxy tool wrappers is supported.
"""

from typing import Optional, cast

from pygls.workspace import Document

from .nodes import (
    XmlAttribute,
    XmlCDATASection,
    XmlComment,
    XmlContent,
    XmlElement,
    XmlProcessingInstruction,
    XmlSyntaxNode,
)
from .document import XmlDocument
from .scanner import XmlScanner
from .types import TokenType


class XmlDocumentParser:
    """Parses a XML document and creates a syntax tree with all the nodes found.

    If the document is incomplete or malformed, the parser will try to recover
    the syntax tree in those cases without altering the original offsets of the nodes."""

    # flake8: noqa: C901
    def parse(self, document: Document) -> XmlDocument:
        """Parses the given text document and returns the resulting syntax tree as
        a XmlDocument.

        This method is a bit too complex, but, since it is a Python translation
        from the Java Eclipse/Lemminx parser, it could be easier to maintain it this way.

        Args:
            document (Document): The XML text document.

        Returns:
            XmlDocument: The resulting syntax tree.
        """
        text = document.source
        text_length = len(text)
        scanner = XmlScanner(text)
        xml_document = XmlDocument(document)
        current: XmlSyntaxNode = xml_document
        attr: Optional[XmlAttribute] = None
        pending_attribute: Optional[str] = None
        last_closed = current
        end_tag_open_offset = -1
        previous_token_was_end_tag_open = False
        token = scanner.scan()
        while token != TokenType.EOS:
            if previous_token_was_end_tag_open:
                previous_token_was_end_tag_open = False
                if token != TokenType.EndTag:
                    # The expected token is not an EndTag, create a fake end tag element
                    self._create_fake_end_tag(end_tag_open_offset, current)

            if token == TokenType.StartTagOpen:
                if not current.is_closed and current.parent:
                    # The next node's parent (current) is not closed at this point
                    # so the node's parent (current) will have its end position updated
                    current.end = scanner.get_token_offset()
                if current.is_closed:
                    # The next node being considered is a child of 'current' and if
                    # 'current' is already closed then 'current' was not updated properly.
                    current = cast(XmlSyntaxNode, current.parent)
                child = XmlElement(scanner.get_token_offset(), scanner.get_token_end())
                child.start_tag_open_offset = scanner.get_token_offset()
                child.parent = current
                current = child

            elif token == TokenType.StartTag:
                element = cast(XmlElement, current)
                element.name = scanner.get_token_text()
                current.end = scanner.get_token_end()

            elif token == TokenType.StartTagClose:
                if current.is_element:
                    current.end = scanner.get_token_end()
                    element = cast(XmlElement, current)
                    element.start_tag_close_offset = scanner.get_token_offset()
                current.end = scanner.get_token_end()

            elif token == TokenType.EndTagOpen:
                end_tag_open_offset = scanner.get_token_offset()
                current.end = scanner.get_token_offset()
                previous_token_was_end_tag_open = True

            elif token == TokenType.EndTag:
                close_tag = scanner.get_token_text()
                node = current
                # eg: <a><b><c></d> will set a,b,c end position to the start of |</d>
                while not (current.is_element and cast(XmlElement, current).is_same_tag(close_tag)) and current.parent:
                    current.end = end_tag_open_offset
                    current = current.parent
                if current != xml_document:
                    current._closed = True
                    if current.is_element:
                        cast(XmlElement, current).end_tag_open_offset = end_tag_open_offset
                    current.end = scanner.get_token_end()
                else:
                    # element open tag not found (ex: <root>) add a fake element which only has an end tag (no start tag)
                    element = XmlElement(scanner.get_token_offset() - 2, scanner.get_token_end())
                    element.end_tag_open_offset = end_tag_open_offset
                    element.name = close_tag
                    element.parent = node
                    current = element

            elif token == TokenType.StartTagSelfClose:
                if current.parent:
                    current._closed = True
                    cast(XmlElement, current).is_self_closed = True
                    current.end = scanner.get_token_end()
                    last_closed = current
                    current = current.parent

            elif token == TokenType.EndTagClose:
                if current.parent:
                    current.end = scanner.get_token_end()
                    last_closed = current
                    if last_closed.is_element:
                        cast(XmlElement, last_closed).end_tag_close_offset = scanner.get_token_offset()
                    current = current.parent

            elif token == TokenType.AttributeName:
                pending_attribute = scanner.get_token_text()
                element = cast(XmlElement, current)
                attr = XmlAttribute(
                    pending_attribute,
                    scanner.get_token_offset(),
                    scanner.get_token_offset() + len(pending_attribute),
                    element,
                )
                element.attributes[pending_attribute] = attr
                current.end = scanner.get_token_end()

            elif token == TokenType.DelimiterAssign:
                if attr:
                    # Sets the value to the '=' position in case there is no AttributeValue
                    attr.set_value(None, scanner.get_token_offset(), scanner.get_token_end())
                    attr.has_delimiter = True

            elif token == TokenType.AttributeValue:
                value = scanner.get_token_text()
                if current.has_attributes and attr:
                    attr.set_value(value, scanner.get_token_offset(), scanner.get_token_offset() + len(value))
                pending_attribute = None
                attr = None
                current.end = scanner.get_token_end()

            elif token == TokenType.CDATATagOpen:
                cdata = XmlCDATASection(scanner.get_token_offset(), text_length)
                cdata.parent = current
                current = cdata

            elif token == TokenType.CDATAContent:
                cdata = cast(XmlCDATASection, current)
                cdata.start_content = scanner.get_token_offset()
                cdata.end_content = scanner.get_token_end()
                current.end = scanner.get_token_end()

            elif token == TokenType.CDATATagClose:
                current.end = scanner.get_token_end()
                current._closed = True
                current = cast(XmlSyntaxNode, current.parent)

            elif token == TokenType.StartCommentTag:
                # In case the tag before the comment tag (current) was not properly
                # closed, current should be set to the root node
                if current.is_closed:
                    current = cast(XmlSyntaxNode, current.parent)
                comment = XmlComment(scanner.get_token_offset(), text_length)
                comment.parent = current
                current = comment

            elif token == TokenType.Content:
                content = XmlContent(scanner.get_token_offset(), scanner.get_token_end())
                content._closed = True
                content.parent = current

            elif token == TokenType.Comment:
                comment = cast(XmlComment, current)
                comment.start_content = scanner.get_token_offset()
                comment.end_content = scanner.get_token_end()

            elif token == TokenType.EndCommentTag:
                current.end = scanner.get_token_end()
                current._closed = True
                current = cast(XmlSyntaxNode, current.parent)

            elif token == TokenType.StartPrologOrPI:
                pi = XmlProcessingInstruction(scanner.get_token_offset(), text_length)
                pi.parent = current
                current = pi

            elif token == TokenType.PIContent:
                pi = cast(XmlProcessingInstruction, current)
                pi.start_content = scanner.get_token_offset()
                pi.end_content = scanner.get_token_end()

            elif token == TokenType.PIEnd:
                current.end = scanner.get_token_end()
                current._closed = True
                current = cast(XmlSyntaxNode, current.parent)

            token = scanner.scan()

        if previous_token_was_end_tag_open:
            previous_token_was_end_tag_open = False
            if token != TokenType.EndTag:
                # The expected token is not an EndTag, create a fake end tag element
                self._create_fake_end_tag(end_tag_open_offset, current)

        while current.parent:
            current.end = text_length
            current = current.parent

        return xml_document

    def _create_fake_end_tag(self, end_tag_open_offset, current):
        # The expected token is not an EndTag, create a fake end tag element
        element = XmlElement(end_tag_open_offset, end_tag_open_offset + 2)
        element.end_tag_open_offset = end_tag_open_offset
        element.parent = current



"""Galaxy Language Server Setup"""

import pathlib
from setuptools import setup, find_packages


PACKAGE_NAME = "galaxy-language-server"
VERSION = "0.5.1"
AUTHOR = "David López"
AUTHOR_EMAIL = "davelopez7391@gmail.com"
DESCRIPTION = "A language server for Galaxy (https://galaxyproject.org) tool wrappers"
KEYWORDS = ["galaxy", "python", "language server"]
LICENSE = "Apache License 2.0"
URL = "https://github.com/davelopez/galaxy-language-server/tree/master/server"

base_directory = pathlib.Path(__file__).parent

long_description = (base_directory / "README.md").read_text(encoding="utf-8")
long_description += (base_directory / "CHANGELOG.md").read_text(encoding="utf-8")

requirements = (base_directory / "requirements.txt").read_text().splitlines()

packages = find_packages(exclude=["*.tests*"])

setup(
    name=PACKAGE_NAME,
    version=VERSION,
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type="text/markdown",
    url=URL,
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    keywords=KEYWORDS,
    license=LICENSE,
    packages=packages,
    include_package_data=True,
    install_requires=requirements,
    python_requires="~=3.8",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "Intended Audience :: Information Technology",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.8",
    ],
)



from typing import List, Optional, Tuple, Union, cast

from anytree import PreOrderIter
from galaxyls.services.tools.constants import (
    ARGUMENT,
    BOOLEAN,
    COMMAND,
    CONFIGFILES,
    DASH,
    DATA,
    ENV_VARIABLES,
    FLOAT,
    INPUTS,
    INTEGER,
    NAME,
    OPTIONAL,
    TEXT,
    TOOL,
    TYPE,
    UNDERSCORE,
)
from galaxyls.services.tools.document import GalaxyToolXmlDocument
from galaxyls.services.tools.generators.snippets import SnippetGenerator
from galaxyls.services.tools.inputs import (
    ConditionalInputNode,
    GalaxyToolInputTree,
    InputNode,
    RepeatInputNode,
    SectionInputNode,
)
from galaxyls.services.xml.nodes import XmlElement
from pygls.lsp.types import Position, Range

ARG_PLACEHOLDER = "TODO_argument"
REPEAT_VAR = "item"
REPEAT_INDEX = "i"


class GalaxyToolCommandSnippetGenerator(SnippetGenerator):
    """This class tries to generate some boilerplate Cheetah code for the command
    section using the information already defined in the inputs and outputs of the tool XML wrapper.
    """

    def __init__(self, tool_document: GalaxyToolXmlDocument, tabSize: int = 4) -> None:
        super().__init__(tool_document, tabSize)

    def _build_snippet(self) -> Tuple[str, bool]:
        """This function tries to generate a code snippet in TextMate format with boilerplate
        Cheetah code extracted from the inputs and outputs of the tool.

        Returns:
            Tuple[str, bool]: The code snippet in TextMate format or an error message if the
            generation failed. The second value of the tuple indicates if it is an error.
        """
        try:
            input_tree = self.expanded_document.analyze_inputs()
            outputs = self.expanded_document.get_outputs()
            result_snippet = self._generate_command_snippet(input_tree, outputs)
            command_section = self.tool_document.find_element(COMMAND)
            if command_section and not command_section.is_self_closed:
                if command_section.get_cdata_section():
                    return (result_snippet, False)
                return (f"<![CDATA[\n\n{result_snippet}\n\n]]>\n", False)
            return (f"<{COMMAND}><![CDATA[\n\n{result_snippet}\n\n]]>\n</{COMMAND}>\n", False)
        except BaseException as ex:
            return (f"Automatic command section generation failed with reason: {ex}", True)

    def _find_snippet_insert_position(self) -> Union[Position, Range]:
        """Returns the position inside the document where command section
        can be inserted.

        If the <command> section does not exists in the file, the best aproximate
        position where the section should be inserted is returned (acording to the IUC
        best practices tag order).

        Returns:
            Position: The position where the command section can be inserted in the document.
        """
        tool = self.tool_document
        section = tool.find_element(COMMAND)
        if section:
            content_range = tool.get_content_range(section)
            if content_range:
                cdata = section.get_cdata_section()
                if cdata:
                    content_range = tool.get_content_range(cdata)
                return content_range.end
            else:  # is self closed <tests/>
                start = tool.get_position_before(section)
                end = tool.get_position_after(section)
                return Range(start=start, end=end)
        else:
            section = tool.find_element(ENV_VARIABLES)
            if section:
                return tool.get_position_before(section)
            section = tool.find_element(CONFIGFILES)
            if section:
                return tool.get_position_before(section)
            section = tool.find_element(INPUTS)
            if section:
                return tool.get_position_before(section)
            section = tool.find_element(TOOL)
            if section:
                return tool.get_content_range(section).end
            return Position(line=0, character=0)

    def _generate_command_snippet(self, input_tree: GalaxyToolInputTree, outputs: List[XmlElement]) -> str:
        snippets = [
            "## Auto-generated command section",
            "## TODO: please review and edit this section as needed",
            "\n## Inputs\n",
        ]
        for node in PreOrderIter(input_tree._root):
            node = cast(InputNode, node)
            if type(node) is ConditionalInputNode:
                indent_level = node.depth - 1
                name_path = self._get_ancestor_name_path(node)
                result = self._conditional_to_cheetah(node, name_path, indent_level)
                snippets.extend(result)
            else:
                result = self._node_to_cheetah(node)
                snippets.extend(result)

        snippets.append("\n## Outputs\n")
        for output in outputs:
            variable = self._output_to_cheetah(output)
            if variable:
                snippets.append(variable)
        return "\n".join(snippets)

    def _get_ancestor_name_path(self, node: InputNode) -> Optional[str]:
        """Returns the names of the ancestor nodes separated by '.'"""
        if node.depth > 1:
            ancestor_names = [node.name for node in node.ancestors[1:]]  # Skip the 'inputs' root node
            result = ".".join(ancestor_names)
            return result

    def _param_to_cheetah(self, param: XmlElement, name_path: Optional[str] = None, indent_level: int = 0) -> str:
        """Converts the given param element to it's Cheetah representation."""
        indentation = self._get_indentation(indent_level)
        argument_attr = param.get_attribute(ARGUMENT)
        name_attr = param.get_attribute(NAME)
        if not name_attr and argument_attr:
            name_attr = argument_attr.lstrip(DASH).replace(DASH, UNDERSCORE)
        type_attr = param.get_attribute(TYPE)
        if name_path:
            name_attr = f"{name_path}.{name_attr}"
        if type_attr == BOOLEAN:
            return f"{indentation}\\${name_attr}"
        if type_attr in [INTEGER, FLOAT]:
            if param.get_attribute(OPTIONAL) == "true":
                return (
                    f"{indentation}#if str(\\${name_attr}):\n"
                    f"{indentation}{self.indent_spaces}{self._get_argument_safe(argument_attr)} \\${name_attr}"
                    f"{indentation}#end if"
                )
        if type_attr in [TEXT, DATA]:
            return f"{indentation}{self._get_argument_safe(argument_attr)} '\\${name_attr}'"
        return f"{indentation}{self._get_argument_safe(argument_attr)} \\${name_attr}"

    def _node_to_cheetah(self, node: InputNode, name_path: Optional[str] = None, indent_level: int = 0) -> List[str]:
        """Converts all the elements of the given node to their Cheetah represetation."""
        result: List[str] = []
        for param in node.params:
            result.append(self._param_to_cheetah(param, name_path, indent_level))
        for repeat in node.repeats:
            result.extend(self._repeat_to_cheetah(repeat, name_path, indent_level=indent_level))
        for section in node.sections:
            result.extend(self._section_to_cheetah(section, name_path, indent_level=indent_level))
        return result

    def _conditional_to_cheetah(
        self, conditional: ConditionalInputNode, name_path: Optional[str] = None, indent_level: int = 0
    ) -> List[str]:
        """Converts the given conditional node to it's Cheetah representation using #if/#elif clauses."""
        indentation = self._get_indentation(indent_level)
        result: List[str] = []
        cond_name = conditional.name
        if name_path:
            cond_name = f"{name_path}.{cond_name}"
        param_name = conditional.option_param.get_attribute(NAME)
        option = conditional.option
        directive = "#elif"
        if conditional.is_first_option:
            directive = "#if"
        result.append(f'{indentation}{directive} str( \\${cond_name}.{param_name} ) == "{option}":')
        result.extend(self._node_to_cheetah(conditional, cond_name, indent_level + 1))
        if conditional.is_last_option:
            result.append(f"{indentation}#end if")
        return result

    def _repeat_to_cheetah(self, repeat: RepeatInputNode, name_path: Optional[str] = None, indent_level: int = 0) -> List[str]:
        """Converts the given repeat node to it's Cheetah representation using the #for directive."""
        indentation = self._get_indentation(indent_level)
        result: List[str] = []
        repeat_name = repeat.element.get_attribute(NAME)
        if name_path:
            repeat_name = f"{name_path}.{repeat_name}"
        var_placeholder = self._get_next_tabstop_with_placeholder(REPEAT_VAR)
        result.append(f"{indentation}#for \\${REPEAT_INDEX}, \\${var_placeholder} in enumerate(\\${repeat_name}):")
        result.extend(self._node_to_cheetah(repeat, var_placeholder, indent_level + 1))
        result.append(f"{indentation}#end for")
        return result

    def _section_to_cheetah(
        self, section: SectionInputNode, name_path: Optional[str] = None, indent_level: int = 0
    ) -> List[str]:
        """Converts the given section node to it's Cheetah representation."""
        result: List[str] = []
        section_name = section.element.get_attribute(NAME)
        if name_path:
            section_name = f"{name_path}.{section_name}"
        result.extend(self._node_to_cheetah(section, section_name, indent_level))
        return result

    def _output_to_cheetah(self, output: XmlElement) -> Optional[str]:
        """Converts the given output element to it's Cheetah representation wrapped in single quotes."""
        name = output.get_attribute(NAME)
        if name:
            return f"'\\${name}'"
        return None

    def _get_argument_safe(self, argument: Optional[str]) -> str:
        """Returns the given argument or a tabstop placeholder if it is None."""
        return argument or self._get_next_tabstop_with_placeholder(ARG_PLACEHOLDER)

    def _get_indentation(self, level: int) -> str:
        """Returns a str with the spaces required for the given indentation level."""
        return self.indent_spaces * level



from typing import Dict, List, Optional

from anytree.search import findall
from pygls.lsp.types import Position, Range
from pygls.workspace import Document

from .nodes import XmlContainerNode, XmlElement, XmlSyntaxNode
from .types import DocumentType, NodeType
from .utils import convert_document_offset_to_position, convert_document_offsets_to_range


class XmlDocument(XmlSyntaxNode):
    """Represents a parsed XML document.

    This is the root of the XML syntax tree.
    """

    def __init__(self, document: Document):
        super().__init__()
        self.document: Document = document
        self.supported_document_types: Dict[str, DocumentType] = {
            "tool": DocumentType.TOOL,
            "macros": DocumentType.MACROS,
        }

    @property
    def node_type(self) -> NodeType:
        """The type of this node."""
        return NodeType.DOCUMENT

    @property
    def is_empty(self) -> bool:
        """True if the document has no root node."""
        return not self.root

    @property
    def root(self) -> Optional[XmlElement]:
        """The root element of the document.

        Normally this would be tool, macros or any other supported
        kind of element."""
        if len(self.children) == 0:
            return None
        try:
            return next(child for child in self.children if type(child) == XmlElement)
        except StopIteration:
            return None

    @property
    def uses_macros(self) -> bool:
        """Indicates if this XML document contains any <expand> element.

        Returns:
            bool: True if the tool contains at least one <expand> elements.
        """
        try:
            found = findall(self.root, filter_=lambda node: node.name == "expand", mincount=1)
            return len(found) > 0
        except BaseException:
            return False

    @property
    def document_type(self) -> DocumentType:
        """The type of this document (if it is supported) or UNKNOWN."""
        if self.root and self.root.name:
            return self.supported_document_types.get(self.root.name, DocumentType.UNKNOWN)
        return DocumentType.UNKNOWN

    @property
    def is_unknown(self) -> bool:
        """Indicates if the document is of unknown type."""
        return self.document_type == DocumentType.UNKNOWN

    @property
    def is_macros_file(self) -> bool:
        """Indicates if the document is a macro definition file."""
        return self.document_type == DocumentType.MACROS

    @property
    def is_tool_file(self) -> bool:
        """Indicates if the document is a tool definition file."""
        return self.document_type == DocumentType.TOOL

    def get_node_at(self, offset: int) -> Optional[XmlSyntaxNode]:
        """Gets the syntax node a the given offset."""
        return self.root.find_node_at(offset)

    def get_content_range(self, element: XmlContainerNode) -> Optional[Range]:
        """Gets the Range positions for the given XML element's content in the document.

        Args:
            element (XmlContainerNode): The XML element to determine it's content range positions.

        Returns:
            Optional[Range]: The range positions for the content of the given XML element.
        """
        start_offset, end_offset = element.get_content_offsets()
        if start_offset < 0 or end_offset < 0:
            return None
        return convert_document_offsets_to_range(self.document, start_offset, end_offset)

    def get_element_name_range(self, element: XmlElement) -> Optional[Range]:
        """Gets the Range positions for the given XML element's name in the document.

        Args:
            element (XmlElement): The XML element to determine it's name range positions.

        Returns:
            Optional[Range]: The range positions for the name of the given XML element.
        """
        start_offset = element.name_start_offset
        end_offset = element.name_end_offset
        if start_offset < 0 or end_offset < 0:
            return None
        return convert_document_offsets_to_range(self.document, start_offset, end_offset)

    def get_full_range(self, node: XmlSyntaxNode) -> Optional[Range]:
        """Gets the Range positions for the given XML node in the document.

        Args:
            node (XmlSyntaxNode): The XML node to determine it's range positions.

        Returns:
            Optional[Range]: The range positions for the entire node.
        """
        if node.start < 0 or node.end < 0:
            return None
        return convert_document_offsets_to_range(self.document, node.start, node.end)

    def get_position_before(self, element: XmlElement) -> Position:
        """Return the position in the document before the given element.

        Args:
            element (XmlElement): The element used to find the position.

        Returns:
            Position: The position just before the element declaration.
        """
        return convert_document_offset_to_position(self.document, element.start)

    def get_position_after(self, element: XmlElement) -> Position:
        """Return the position in the document after the given element.

        Args:
            element (XmlElement): The element used to find the position.

        Returns:
            Position: The position just after the element declaration.
        """
        if element.is_self_closed:
            return convert_document_offset_to_position(self.document, element.end)
        return convert_document_offset_to_position(self.document, element.end_offset)

    def find_all_elements_with_name(self, name: str) -> List[XmlElement]:
        """Returns a list with all the elements contained in the document matching the given name."""
        if self.root:
            found = findall(self.root, filter_=lambda node: isinstance(node, XmlElement) and node.name == name)
            return list(found)
        return []

    def get_text_between_offsets(self, start: int, end: int) -> str:
        """Gets the text content between the start and end offsets."""
        return self.document.source[start:end]



""" Type definitions for XSD processing.
"""

from typing import Dict, List, Optional, cast

from anytree import NodeMixin, RenderTree, Resolver, ResolverError, findall
from lxml import etree
from pygls.lsp.types import MarkupContent, MarkupKind

from .constants import MSG_NO_DOCUMENTATION_AVAILABLE


class XsdBase:
    """Base class that encapsulates an element from the XSD schema.

    It contains common information and functionality shared by
    XML nodes and attributes.
    """

    def __init__(self, name: str, element: Optional[etree._Element]):
        super(XsdBase, self).__init__()
        self.name: str = name
        self.xsd_element: Optional[etree._Element] = element
        self.xsd_type: Optional[etree._Element] = None

    def __repr__(self) -> str:
        return self.name

    def get_doc(self, lang: str = "en") -> MarkupContent:
        """Gets the Markdown documentation associated with this element
        from the XSD schema.

        If there is no documentation in the schema for the element,
        a message indicating this will be returned instead.

        Args:
            lang (str, optional): The language code of the documentation
            to retrieve. Defaults to "en" (English).

        Returns:
            [str]: The documentation text or a message indicating
            there is no documentation.
        """
        doc = self._get_doc_text_of_element(self.xsd_element, lang) or self._get_doc_text_of_element(self.xsd_type, lang)
        if doc:
            return MarkupContent(kind=MarkupKind.Markdown, value=doc)
        return MarkupContent(kind=MarkupKind.Markdown, value=MSG_NO_DOCUMENTATION_AVAILABLE)

    def _get_doc_text_of_element(self, element: Optional[etree._Element], lang: str = "en") -> str:
        try:
            doc_annotation = element.xpath(
                "./xs:annotation/xs:documentation[@xml:lang=$lang]/text()",
                namespaces=element.nsmap,
                lang=lang,
            )
            return cast(List[str], doc_annotation)[0].strip()
        except BaseException:
            return ""


class XsdAttribute(XsdBase):
    """Represents an attribute in an XML tag.

    It contains information about the attribute extracted
    from the XSD schema.

    Args:
        XsdBase: Inherits base functionality from XsdBase.
    """

    def __init__(
        self,
        name: str,
        element: Optional[etree._Element],
        type_name: Optional[str] = None,
        is_required: bool = False,
    ):
        super(XsdAttribute, self).__init__(name, element)
        self.type_name: Optional[str] = type_name
        self.is_required: bool = is_required
        self.enumeration: List[str] = []


class XsdNode(XsdBase, NodeMixin):
    """Represents a particular XML tag.

    The node contains information extracted from the XSD schema
    about the XML tag, possible descendants, attributes, etc.

    Args:
        XsdBase: Inherits base functionality from XsdBase.
        NodeMixin: Inherits tree node functionality from NodeMixin.
    """

    def __init__(self, name: str, element: Optional[etree._Element] = None, parent: Optional[NodeMixin] = None):
        super(XsdNode, self).__init__(name, element)
        self.parent: Optional[NodeMixin] = parent
        self.attributes: Dict[str, XsdAttribute] = {}
        self.min_occurs: int = 1  # required by default
        self.max_occurs: int = -1  # unbounded by default

    def render(self) -> str:
        """Gets an ascii representation of this node.

        Returns:
            str: An ascii representation of the node.
        """
        return str(RenderTree(self).by_attr(lambda node: f"[{node.name}] {' '.join(node.attributes)}"))


class XsdTree:
    """Represents a tree structure containing the important
    XSD information for all the elements and attributes.
    """

    def __init__(self, root: XsdNode):
        self.root: XsdNode = root
        self.node_resolver = Resolver("name")
        self.expand_element = self._build_expand_element()

    def find_node_by_stack(self, node_stack: List[str]) -> Optional[XsdNode]:
        """Finds the node definition in the tree that matches the given stack of tags.

        Args:
            node_stack (List[str]): The stack of tag names composing a tree branch.
            Like: ['root', 'node', 'subnode', 'leaf']

        Returns:
            Optional[XsdNode]: The node definition matching the leaf in the path or
            None if the node could not be found.
        """
        result_node = None
        if node_stack:
            try:
                path = self._get_path_from_stack(node_stack)
                if path.endswith(self.expand_element.name):
                    return self.expand_element
                result_node = self.node_resolver.get(self.root, path)
            except ResolverError as e:
                print(e)
        return result_node

    def render(self) -> str:
        """Gets an ascii representation of this tree.

        Returns:
            str: An ascii representation of the tree.
        """
        return self.root.render()

    def _get_path_from_stack(self, node_stack: List[str]) -> str:
        if node_stack[0] == self.root.name:
            node_stack[0] = "."
        return "/".join(node_stack)

    def _build_expand_element(self) -> XsdNode:
        """Creates a XsdNode representing the <expand> element and it's attributes.
        The <expand> element is a special element that is not defined in the XSD schema but allows to use macros in the tool.

        Returns:
            XsdNode: The node with information about the special <expand> element.
        """
        expand_node = XsdNode("expand", None)
        attr_name = "macro"
        attr = XsdAttribute(attr_name, None, type_name=None, is_required=True)
        expand_node.attributes[attr_name] = attr
        return expand_node

    def find_node_by_name(self, name: str) -> Optional[XsdNode]:
        if name == self.expand_element.name:
            return self.expand_element

        result = findall(
            self.root,
            filter_=lambda node: node.name == name,
        )
        if len(result) > 0:
            return result[0]



"""This module provides a service to determine position context inside an XML document."""

from typing import List, Optional, cast

from pygls.lsp.types import Range
from pygls.workspace import Position

from galaxyls.services.tools.constants import MACROS
from galaxyls.services.xml.constants import UNDEFINED_OFFSET
from galaxyls.services.xml.document import XmlDocument
from galaxyls.services.xml.nodes import XmlElement, XmlSyntaxNode
from galaxyls.services.xml.types import NodeType
from galaxyls.services.xml.utils import convert_document_offsets_to_range
from galaxyls.services.xsd.types import XsdNode, XsdTree


class XmlContext:
    """Represents the context at a given XML document position.

    It provides information about the token under the cursor and
    the XSD node definition associated.
    """

    def __init__(
        self,
        xml_document: XmlDocument,
        xsd_node: Optional[XsdNode],
        node: Optional[XmlSyntaxNode] = None,
        line_text: str = "",
        position: Optional[Position] = None,
        offset: int = UNDEFINED_OFFSET,
    ):
        self.xml_document = xml_document
        self._xsd_node = xsd_node
        self._node = node
        self._line_text = line_text
        self._position = position
        self._offset = offset

    @property
    def node(self) -> Optional[XmlSyntaxNode]:
        """The syntax node at the context position."""
        return self._node

    @property
    def position(self) -> Optional[Position]:
        """The context position (line and character) inside de Document."""
        return self._position

    @property
    def offset(self) -> int:
        """The character offset inside de Document."""
        return self._offset

    @property
    def line_text(self) -> str:
        """The text contents of the document line at the context position."""
        return self._line_text

    @property
    def characted_at_position(self) -> Optional[str]:
        """The character at the context position."""
        try:
            return self._line_text[self._position.character]
        except (IndexError, AttributeError):
            return None

    @property
    def xsd_element(self) -> Optional[XsdNode]:
        """The XSD element associated with the token in context."""
        return self._xsd_node

    @property
    def is_empty(self) -> bool:
        """Indicates if the document is empty and no context can be determined."""
        return not self._node

    @property
    def is_root(self) -> bool:
        """Indicates if the element at context is the root element."""
        return self._node is not None and len(self._node.ancestors) == 1

    @property
    def is_tag(self) -> bool:
        """Indicates if the token in context is a tag."""
        return self._node is not None and self._node.node_type == NodeType.ELEMENT

    @property
    def is_tag_name(self) -> bool:
        """Indicates if the token in context is a start tag name."""
        return (
            self.is_tag
            and cast(XmlElement, self._node).name_start_offset <= self._offset <= cast(XmlElement, self._node).name_end_offset
        )

    @property
    def is_attribute(self) -> bool:
        """Indicates if the token in context is an attribute."""
        return self._node is not None and self._node.is_attribute

    @property
    def is_attribute_key(self) -> bool:
        """Indicates if the token in context is an attribute key."""
        return self._node is not None and self._node.node_type == NodeType.ATTRIBUTE_KEY

    @property
    def is_attribute_value(self) -> bool:
        """Indicates if the token in context is an attribute value."""
        return self._node is not None and self._node.node_type == NodeType.ATTRIBUTE_VALUE

    @property
    def is_inside_attribute_value(self) -> bool:
        """Indicates if the token in context is an attribute value."""
        return self.is_attribute_value and self.offset > self.node.start and self.offset < self.node.end

    @property
    def is_attribute_end(self) -> bool:
        """Indicates that the context position is at the ending quote character of an attribute.

        Example: <tag attribute="value["] <- The context position is at " """
        return self.is_attribute_value and self._offset == self._node.end - 1

    @property
    def attribute_name(self) -> Optional[str]:
        """The name of the attribute if the context is an attribute or None."""
        return self._node and self._node.get_attribute_name()

    @property
    def is_content(self) -> bool:
        """Indicates if the token in context is within a content or CDATA block."""
        return self._node is not None and self._node.node_type in [NodeType.CONTENT, NodeType.CDATA_SECTION]

    @property
    def is_closing_tag(self) -> bool:
        """Indicates if the token in context is a closing tag."""
        return self._node is not None and self._node.is_at_closing_tag(self._offset)

    @property
    def is_at_end(self) -> bool:
        """Indicates if the context is at the last character of the node."""
        return self._node is not None and self._node.end == self._offset

    @property
    def stack(self) -> List[str]:
        """The list of XML tag names from the root to the token in context."""
        if not self._node:
            return []
        return self._node.stack

    def has_reached_max_occurs(self, node: XsdNode) -> bool:
        """Checks if the given node has reached the maximum number
        of occurrences.

        Args:
            child (XsdNode): The node to check.

        Returns:
            bool: True if the node has reached the maximum number
            of occurrences permitted.
        """
        if node.max_occurs < 0:
            return False
        target = self._node.parent or self._node
        if target:
            existing_count = sum(1 for child_node in target.children if child_node.name == node.name)
            return existing_count >= node.max_occurs
        return False

    def is_valid_tag(self) -> bool:
        """Indicates"""
        if self.is_tag and self._xsd_node is not None:
            return self._node.name == self._xsd_node.name or self._node.name in [node.name for node in self._xsd_node.children]
        return False


class XmlContextService:
    """This service provides information about the XML context at
    a specific position of the document.
    """

    def __init__(self, xsd_tree: XsdTree):
        self.xsd_tree = xsd_tree

    def get_xml_context(self, xml_document: XmlDocument, position: Position) -> XmlContext:
        """Gets the XML context at a given position inside the document.

        Args:
            document (Document): The current document.
            position (Position): The position inside de document.

        Returns:
            XmlContext: The resulting context with the current node
            definition and other information. If the context can not be
            determined, the default context with no information is returned.
        """
        offset = xml_document.document.offset_at_position(position)

        if xml_document.is_empty:
            return XmlContext(xml_document, self.xsd_tree.root, node=None)
        node = xml_document.get_node_at(offset)
        xsd_node = self.find_matching_xsd_element(node, self.xsd_tree)
        line_text = xml_document.document.lines[position.line]
        context = XmlContext(xml_document, xsd_node, node, line_text, position, offset)
        return context

    def find_matching_xsd_element(self, node: Optional[XmlSyntaxNode], xsd_tree: XsdTree) -> Optional[XsdNode]:
        """Finds the xsd element in the XSD tree that matches the xml element associated with the given syntax node.
        If there is no matching node, the root (tool) xsd node is always returned.

        Args:
            node (XmlSyntaxNode): The syntax node to match.
            xsd_tree (XsdTree): The XSD tree definition.

        Returns:
            XsdNode: The matching xsd node or None if there is no matching.
        """
        if node:
            node_stack = node.stack
            if len(node_stack) > 0 and MACROS in node_stack:
                return xsd_tree.find_node_by_name(node_stack[-1])
            xsd_node = xsd_tree.find_node_by_stack(node_stack)
            if xsd_node is None:
                xsd_node = xsd_tree.find_node_by_stack(node_stack[:-1])
            return xsd_node

    def get_range_for_context(self, xml_document: XmlDocument, context: XmlContext) -> Range:
        start_offset, end_offset = context.node.get_offsets(context.offset)
        return convert_document_offsets_to_range(xml_document.document, start_offset, end_offset)



from enum import Enum

from pydantic import BaseModel


def to_camel(string: str) -> str:
    pascal = "".join(word.capitalize() for word in string.split("_"))
    camel = pascal[0].lower() + pascal[1:]
    return camel


class CompletionMode(str, Enum):
    AUTO = "auto"
    INVOKE = "invoke"
    DISABLED = "disabled"


class ConfigModel(BaseModel):
    class Config:
        alias_generator = to_camel
        allow_population_by_field_name = True


class CompletionConfig(ConfigModel):
    """Auto-completion feature configuration."""

    mode: CompletionMode = CompletionMode.AUTO
    auto_close_tags: bool = True


class GalaxyToolsConfiguration(ConfigModel):
    """Galaxy Language Server general configuration."""

    completion: CompletionConfig = CompletionConfig()
